# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.vx11
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements a VXI11Driver based class to control VXI11 instruments

    Loosely based on Python Sun RPC Demo and Alex Forencich python-vx11

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import enum
import socket

from lantz import errors
from lantz.drivers.legacy import rpc
from lantz import Driver


# VXI-11 RPC constants

# Device async
DEVICE_ASYNC_PROG = 0x0607b0
DEVICE_ASYNC_VERS = 1
DEVICE_ABORT      = 1

# Device core
DEVICE_CORE_PROG  = 0x0607af
DEVICE_CORE_VERS  = 1
CREATE_LINK       = 10
DEVICE_WRITE      = 11
DEVICE_READ       = 12
DEVICE_READSTB    = 13
DEVICE_TRIGGER    = 14
DEVICE_CLEAR      = 15
DEVICE_REMOTE     = 16
DEVICE_LOCAL      = 17
DEVICE_LOCK       = 18
DEVICE_UNLOCK     = 19
DEVICE_ENABLE_SRQ = 20
DEVICE_DOCMD      = 22
DESTROY_LINK      = 23
CREATE_INTR_CHAN  = 25
DESTROY_INTR_CHAN = 26

# Device intr
DEVICE_INTR_PROG  = 0x0607b1
DEVICE_INTR_VERS  = 1
DEVICE_INTR_SRQ   = 30

# Error states
class ERROR_CODES(enum.IntEnum):
    NO_ERROR = 0
    SYNTAX_ERROR = 1
    DEVICE_NOT_ACCESSIBLE = 3
    INVALID_LINK_IDENTIFIER = 4
    PARAMETER_ERROR = 5
    CHANNEL_NOT_ESTABLISHED = 6
    OPERATION_NOT_SUPPORTED = 8
    OUT_OF_RESOURCES = 9
    DEVICE_LOCKED_BY_ANOTHER_LINK = 11
    NO_LOCK_HELD_BY_THIS_LINK = 12
    IO_TIMEOUT = 15
    IO_ERROR = 17
    ABORT = 23
    CHANNEL_ALREADY_ESTABLISHED = 29

# Flags
OP_FLAG_WAIT_BLOCK = 1
OP_FLAG_END = 8
OP_FLAG_TERMCHAR_SET = 128

RX_REQCNT = 1
RX_CHR = 2
RX_END = 4

# Exceptions
class Vxi11Error(InstrumentError):
    pass


class Vxi11Packer(rpc.Packer):
    def pack_device_link(self, link):
        self.pack_int(link)
    
    def pack_create_link_parms(self, params):
        id, lock_device, lock_timeout, device = params
        self.pack_int(id)
        self.pack_bool(lock_device)
        self.pack_uint(lock_timeout)
        self.pack_string(device)
    
    def pack_device_write_parms(self, params):
        link, io_timeout, lock_timeout, flags, data = params
        self.pack_int(link)
        self.pack_uint(io_timeout)
        self.pack_uint(lock_timeout)
        self.pack_int(flags)
        self.pack_opaque(data)
    
    def pack_device_read_parms(self, params):
        link, request_size, io_timeout, lock_timeout, flags, term_char = params
        self.pack_int(link)
        self.pack_uint(request_size)
        self.pack_uint(io_timeout)
        self.pack_uint(lock_timeout)
        self.pack_int(flags)
        self.pack_int(term_char)
    
    def pack_device_generic_parms(self, params):
        link, flags, lock_timeout, io_timeout = params
        self.pack_int(link)
        self.pack_int(flags)
        self.pack_uint(lock_timeout)
        self.pack_uint(io_timeout)
    
    def pack_device_remote_func_parms(self, params):
        host_addr, host_port, prog_num, prog_vers, prog_family = params
        self.pack_uint(host_addr)
        self.pack_uint(host_port)
        self.pack_uint(prog_num)
        self.pack_uint(prog_vers)
        self.pack_int(prog_family)
    
    def pack_device_enable_srq_parms(self, params):
        link, enable, handle = params
        self.pack_int(link)
        self.pack_bool(enable)
        if len(handle) > 40:
            raise Vxi11Error("array length too long")
        self.pack_opaque(handle)
    
    def pack_device_lock_parms(self, params):
        link, flags, lock_timeout = params
        self.pack_int(link)
        self.pack_int(flags)
        self.pack_uint(lock_timeout)
    
    def pack_device_docmd_parms(self, params):
        link, flags, io_timeout, lock_timeout, cmd, network_order, datasize, data_in = params
        self.pack_int(link)
        self.pack_int(flags)
        self.pack_uint(io_timeout)
        self.pack_uint(lock_timeout)
        self.pack_int(cmd)
        self.pack_bool(network_order)
        self.pack_int(datasize)
        self.pack_opaque(data_in)


class Vxi11Unpacker(rpc.Unpacker):
    def unpack_device_link(self):
        return self.unpack_int()
    
    def unpack_device_error(self):
        return self.unpack_int()
    
    def unpack_create_link_resp(self):
        error = self.unpack_int()
        link = self.unpack_int()
        abort_port = self.unpack_uint()
        max_recv_size = self.unpack_uint()
        return error, link, abort_port, max_recv_size
    
    def unpack_device_write_resp(self):
        error = self.unpack_int()
        size = self.unpack_uint()
        return error, size
    
    def unpack_device_read_resp(self):
        error = self.unpack_int()
        reason = self.unpack_int()
        data = self.unpack_opaque()
        return error, reason, data
    
    def unpack_device_read_stb_resp(self):
        error = self.unpack_int()
        stb = self.unpack_uint()
        return error, stb
    
    def unpack_device_docmd_resp(self):
        error = self.unpack_int()
        data_out = self.unpack_opaque()
        return error, data_out


class CoreClient(rpc.TCPClient):

    def __init__(self, host):
        self.packer = Vxi11Packer()
        self.unpacker = Vxi11Unpacker('')
        super().__init__(host, DEVICE_CORE_PROG, DEVICE_CORE_VERS)
        # I have disabled the creation of the socket by overriding connect.
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def connect(self):
        # Overrides connect method from parent class to avoid automatic
        # opening of connection.
        pass

    def create_link(self, id, lock_device, lock_timeout, name):
        params = (id, lock_device, lock_timeout, name)
        return self.make_call(CREATE_LINK, params,
                              self.packer.pack_create_link_parms,
                              self.unpacker.unpack_create_link_resp)

    def device_write(self, link, io_timeout, lock_timeout, flags, data):
        params = (link, io_timeout, lock_timeout, flags, data)
        return self.make_call(DEVICE_WRITE, params,
                              self.packer.pack_device_write_parms,
                              self.unpacker.unpack_device_write_resp)

    def device_read(self, link, request_size, io_timeout, lock_timeout, flags, term_char):
        params = (link, request_size, io_timeout, lock_timeout, flags, term_char)
        return self.make_call(DEVICE_READ, params,
                              self.packer.pack_device_read_parms,
                              self.unpacker.unpack_device_read_resp)

    def device_read_stb(self, link, flags, lock_timeout, io_timeout):
        params = (link, flags, lock_timeout, io_timeout)
        return self.make_call(DEVICE_READSTB, params,
                              self.packer.pack_device_generic_parms,
                              self.unpacker.unpack_device_read_stb_resp)

    def device_trigger(self, link, flags, lock_timeout, io_timeout):
        params = (link, flags, lock_timeout, io_timeout)
        return self.make_call(DEVICE_TRIGGER, params,
                              self.packer.pack_device_generic_parms,
                              self.unpacker.unpack_device_error)

    def device_clear(self, link, flags, lock_timeout, io_timeout):
        params = (link, flags, lock_timeout, io_timeout)
        return self.make_call(DEVICE_CLEAR, params,
                              self.packer.pack_device_generic_parms,
                              self.unpacker.unpack_device_error)

    def device_remote(self, link, flags, lock_timeout, io_timeout):
        params = (link, flags, lock_timeout, io_timeout)
        return self.make_call(DEVICE_REMOTE, params,
                              self.packer.pack_device_generic_parms,
                              self.unpacker.unpack_device_error)

    def device_local(self, link, flags, lock_timeout, io_timeout):
        params = (link, flags, lock_timeout, io_timeout)
        return self.make_call(DEVICE_LOCAL, params,
                              self.packer.pack_device_generic_parms,
                              self.unpacker.unpack_device_error)

    def device_lock(self, link, flags, lock_timeout):
        params = (link, flags, lock_timeout)
        return self.make_call(DEVICE_LOCK, params,
                              self.packer.pack_device_lock_parms,
                              self.unpacker.unpack_device_error)

    def device_unlock(self, link):
        return self.make_call(DEVICE_UNLOCK, link,
                              self.packer.pack_device_link,
                              self.unpacker.unpack_device_error)

    def device_enable_srq(self, link, enable, handle):
        params = (link, enable, handle)
        return self.make_call(DEVICE_ENABLE_SRQ, params,
                              self.packer.pack_device_enable_srq_parms,
                              self.unpacker.unpack_device_error)

    def device_docmd(self, link, flags, io_timeout, lock_timeout, cmd, network_order, datasize, data_in):
        params = (link, flags, io_timeout, lock_timeout, cmd, network_order, datasize, data_in)
        return self.make_call(DEVICE_DOCMD, params,
                              self.packer.pack_device_docmd_parms,
                              self.unpacker.unpack_device_docmd_resp)

    def destroy_link(self, link):
        return self.make_call(DESTROY_LINK, link,
                              self.packer.pack_device_link,
                              self.unpacker.unpack_device_error)
    
    def create_intr_chan(self, host_addr, host_port, prog_num, prog_vers, prog_family):
        params = (host_addr, host_port, prog_num, prog_vers, prog_family)
        return self.make_call(CREATE_INTR_CHAN, params,
                              self.packer.pack_device_docmd_parms,
                              self.unpacker.unpack_device_error)
    
    def destroy_intr_chan(self):
        return self.make_call(DESTROY_INTR_CHAN, None,
                              None,
                              self.unpacker.unpack_device_error)


class Vxi11Driver(Driver):
    """VXI-11 instrument interface client.
    We do not inherit from TCPDriver because the RPC implementation has its own socket.
    """

    def __init__(self, host='localhost', *args, **kwargs):
        super().__init__(host, *args, **kwargs)

        self.socket = None
        self.lock_timeout = 10000
        self.io_timeout = 10000
        self.client = CoreClient(host)
        self.link = None

    def initialize(self):
        """Open connection to VXI-11 instrument
        """
        super().initialize()
        self.socket = self.client.sock
        self.socket.connect((self.client.sock, self.client.port))

        error, link, abort_port, max_recv_size = self.client.create_link(self.client_id, 0, self.lock_timeout, self.name.encode(self.ENCODING))
        
        if error:
            raise Vxi11Error("error creating link: %d" % error)
        
        self.link = link
        self.RECV_CHUNK = max_recv_size
        
    def finalize(self):
        """Close connection
        """
        self.client.destroy_link(self.link)
        self.client.close()
        super().finalize()

    def write_raw(self, data):
        """Write binary data to instrument
        """

        if self.term_char is not None:
            flags = OP_FLAG_TERMCHAR_SET
            term_char = str(self.term_char).encode('utf-8')[0]
            data += term_char
        
        flags = 0
        
        num = len(data)
        
        offset = 0
        
        while num > 0:
            if num <= self.RECV_CHUNK:
                flags |= OP_FLAG_END
            
            block = data[offset:offset+self.RECV_CHUNK]
            
            error, size = self.client.device_write(self.link, self.io_timeout, self.lock_timeout, flags, block)
            
            if error:
                raise Vxi11Error("error writing data: %d" % error)
            elif size < len(block):
                raise Vxi11Error("did not write complete block")
            
            offset += size
            num -= size

    def read_raw(self, num=-1):
        """Read binary data from instrument
        """

        read_len = self.RECV_CHUNK
        if 0 < num < self.RECV_CHUNK:
            read_len = num
        
        flags = 0
        reason = 0
        
        term_char = 0
        
        if self.term_char is not None:
            flags = OP_FLAG_TERMCHAR_SET
            term_char = str(self.term_char).encode('utf-8')[0]
        
        read_data = b''
        
        while reason & (RX_END | RX_CHR) == 0:
            error, reason, data = self.client.device_read(self.link, read_len, self.io_timeout, self.lock_timeout, flags, term_char)
            
            if error:
                raise Vxi11Error("error reading data: %d" % error)
            
            read_data += data
            
            if num > 0:
                num = num - len(data)
                if num <= 0:
                    break
                if num < read_len:
                    read_len = num
            
        return read_data

    def read_stb(self):
        """Read status byte
        """

        flags = 0
        
        error, stb = self.client.device_read_stb(self.link, flags, self.lock_timeout, self.io_timeout)
        
        if error:
            raise Vxi11Error("error reading status: %d" % error)
        
        return stb
    
    def trigger(self):
        """Send trigger command.
        """

        flags = 0
        
        error = self.client.device_trigger(self.link, flags, self.lock_timeout, self.io_timeout)
        
        if error:
            raise Vxi11Error("error triggering: %d" % error)
    
    def clear(self):
        """Send clear command.
        """

        flags = 0

        error = self.client.device_clear(self.link, flags, self.lock_timeout, self.io_timeout)
        
        if error:
            raise Vxi11Error("error clearing: %d" % error)
    
    def remote(self):
        """Send remote command.
        """

        flags = 0
        
        error = self.client.device_remote(self.link, flags, self.lock_timeout, self.io_timeout)
        
        if error:
            raise Vxi11Error("error remote: %d" % error)
    
    def local(self):
        """Send local command
        """
        
        flags = 0
        
        error = self.client.device_local(self.link, flags, self.lock_timeout, self.io_timeout)
        
        if error:
            raise Vxi11Error("error local: %d" % error)
    
    def lock(self):
        """Send lock command.
        """

        flags = 0
        
        error = self.client.device_lock(self.link, flags, self.lock_timeout)
        
        if error:
            raise Vxi11Error("error locking: %d" % error)
    
    def unlock(self):
        """Send unlock command.
        """

        flags = 0
        
        error = self.client.device_unlock(self.link)
        
        if error:
            raise Vxi11Error("error unlocking: %d" % error)






# -*- coding: utf-8 -*-
"""
    lantz.drivers.kentech.hri
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the driver for Kentech High Repetition Rate Image Intensifier
    revisions 1 and 2.


    Implementation Notes
    --------------------

    The set of commands is cumbersome and inconsistent. Moreover, each revision
    introduces backward incompatible changes. The Lantz driver abstracts those
    differences.

    Sources::

        - LaVision PicoStar HR12
        - HRI Commands obtained from Kentech
        - HRI.cl from DaVis
        - Lantz reverse engineering team


    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.

"""

from lantz import Feat, Action
from lantz import errors
from lantz import MessageBasedDriver

def between(s, before, after):
    ndx1 = s.index(before)
    ndx2 = s.index(after)
    return s[ndx1+len(before):ndx2]


class HRI(MessageBasedDriver):
    """Kentech High Repetition Rate Image Intensifier.
    """

    DEFAULTS = {'COMMON': {'write_termination': '\r',
                           'read_termination': '\n'}}

    def query(self, command, *, send_args=(None, None), recv_args=(None, None)):
        """Send query to the instrument and return the answer.
        Set remote mode if needed.
        """
        if command and not self.recall('remote'):
            self.log_info('Setting Remote.')
            self.remote = True
        return super().query(command, send_args=send_args, recv_args=recv_args)

    def query_expect(self, command, read_termination=None, expected='ok'):
        """Send a query and check that the answer contains the string.

        :type command: str
        :type read_termination: str | None
        :type expected: str | None
        """
        if command and not self.recall('remote'):
            self.log_info('Setting Remote.')
            self.remote = True
        self.resource.write(command)
        ans = self.read(read_termination)
        if expected and not expected in ans:
            raise errors.InstrumentError("'{}' not in '{}'".format(expected, ans))
        return ans

    @Action()
    def clear(self):
        """Clear the buffer.
        """
        self.write('\r\r')

    @Feat(None, values={True, False})
    def remote(self, value):
        """Remote or local.
        """
        if value:
            #self.query_expect('', None, None)
            self.query_expect('\r', expected=None)
            self.read()
        else:
            return self.query_expect('LOCAL', chr(0), None)

    @Feat(read_once=True)
    def revision(self):
        """Revision.
        """
        ans = self.query_expect('.REV', expected=None)
        print(ans)
        if 'UNDEFINED' in ans:
            ans = '1.0'
        else:
            ans = self.read()
            ans = ans.split()[1]
            return ans

    @Feat(None, values={'ecl': 'ECLTRIG', 'ttl': 'TTLTRIG'})
    def trigger_logic(self, value):
        """Trigger logic.
        """
        self.query_expect(value)

    @Feat(None, values={'high': 'HITRIG', '50ohm': '50TRIG}'})
    def trigger_ttl_termination(self, value):
        """Trigger termination for TTL logic (for ECL is fixed to 50 ohm).
        """
        if self.recall('trigger_type') == 'ecl':
            raise errors.InstrumentError('Level triggering only with ECL')
        self.query_expect(value)

    @Feat(None, values={'rising': '+VETRIG', 'falling': '-VETRIG}'})
    def trigger_edge(self, value):
        """Trigger on rising or falling edge.
        """
        self.query_expect(value)

    @Feat(None, values={'level': 'LVLTRIG', 'log': 'LOGTRIG}'})
    def trigger_ecl_mode(self, value):
        """Trigger mode for ECL logic.
        """
        if self.recall('trigger_type') == 'ttl':
            raise errors.InstrumentError('Level triggering only with ECL')

        self.query_expect(value)

    @Feat(units='centivolt', limits=(-40, 40, 1))
    def trigger_ecl_level(self):
        """Trigger level for ECL logic, mode level.
        """
        if self.revision >= 2.0:
            ans = self.query_expect('THRESHV ?')
            ans = between(ans, 'THRESHV ?', 'ok')
            return float(ans.strip())
        else:
            ans = self.query_expect('THRESHV @ .')[8:]
            try:
                pos = ans.index('.')
            except ValueError:
                raise errors.InstrumentError('Unsupported operation.')
            return float(ans[pos+2:pos+7])

    @trigger_ecl_level.setter
    def trigger_ecl_level(self, value):
        if self.revision >= 2.0:
            self.query_expect('{:d} !THRESH'.format(value))
        else:
            value = 40 * value + 2000.0
            self.query_expect('{:d} THRESH ! TRIG+RF>HW'.format(value))

    @Feat(units='volt', limits=(-50, 50))
    def clamp_voltage(self):
        """Most negative value of the gate pulse.
        """
        if self.revision >= 2.0:
            ans = self.query_expect('CLAMP ?')
            ans = between(ans, 'CLAMP ?', 'ok').strip()
            return float(ans)
        else:
            ans = self.query_expect('CLAMP @ .')
            try:
                pos = ans.index('.')
            except ValueError:
                raise errors.InstrumentError('Unsupported operation.')
            return float(ans[pos+2:pos+7]) / 10.0

    @clamp_voltage.setter
    def clamp_voltage(self, value):
        average = self.recall('average_voltage')
        mn, mx = average - Q_(60, volt), average
        if mn < value < mx:
            raise ValueError('Invalid clamp voltage. Not in range {}-{}'.format(mn, mx))
        self.query_expect('{:d} CLAMP ! CLAMP>HW'.format(value * 10))

    @Feat(units='volt', limits=(-50, 50))
    def average_voltage(self):
        """Cathode potential bias with respect of MCP.
        """
        if self.revision >= 2.0:
            ans = self.query_expect('AVE ?')
            ans = between(ans, 'AVE ?', 'ok')
            return float(ans.strip()) / 10.
        else:
            ans = self.query_expect('THRESHV @ .')[8:]
            try:
                pos = ans.index('.')
            except ValueError:
                raise errors.InstrumentError('Unsupported operation.')
            return float(ans[pos+2:pos+7]) / 10.

    @average_voltage.setter
    def average_voltage(self, value):
        self.query_expect('{:d} AVE ! AVE>HW'.format(value * 10))

    @Feat()
    def status(self):
        """Get status.
        """
        return self.query_expect(".STATUS", chr(0))

    @Feat(None, units='volt', limits=(0, 1700))
    def mcp(self, value):
        """MCP Voltage.
        """
        if self.revision >= '2.0':
            return self.query_expect('{} !MCP'.format(value))
        else:
            return self.query_expect('{} !MCPVOLTS'.format(value))

    @Feat(None, values={'inhibit': 0, 'rf': 21, 'ldc': 22, 'hdc': 23, 'dc': 24,
                     'user1': 25, 'user2': 26, 'user3': 27, 'user4': 28})
    def mode(self, mode):
        """Gain modulation mode.

        HRI Machine Modes and Mode Indices
        None     Mode
        0        INHIBIT
        2-10     COMB modes 200 ps to 1 ns inclusive (High rate operation)
        11-20    COMB modes 100 ps to 3 ns inclusive (Low rate (+GOI) operation)
        21       RF
        22       logic low duty cycle (LDC)
        23       logic high duty cycle
        24       DC
        25-28    user modes 1 to 4
        """
        #TODO: Modes [11-20] not available in rev < 2.0
        return self.query_expect("{} !MODE".format(mode))

    @Feat(None)
    def rfgain(self, value):
        """RF Gain.
        """
        return self.query("{} !RFGAIN".format(value))

    @Feat()
    def temperature(self):
        """Temperature.
        """
        if self.revision == 2.0:
            return self.query("@TEMP .")
        return 0

    @Feat(None, values={True, False})
    def enabled(self, value):
        """MCP Enabled
        """
        if self.revision < 2:
            if value:
                self.query_expect('+M')
            else:
                self.query_expect('-M')
        else:
            if value:
                self.mode = self.__dict__.get('_last_mode', 21)
            else:
                self._last_mode = self.recall('mode')
                self.mode = 0


if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test Kentech HRI')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False, help='Show interactive GUI')
    parser.add_argument('-p', '--port', type=str, default='17',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantz.log.log_to_socket(lantz.log.DEBUG)
    with HRI.from_serial_port(args.port, baudrate=9600) as inst:
        if args.interactive:
            from lantz.ui.app import start_test_app
            start_test_app(inst)
        else:
            #inst.clear()
            inst.remote = True
            print(inst.revision)
            inst.mode = "inhibit"
            inst.mcp = 350
            inst.rfgain = 99
            #print(inst.status)
            inst.mode = "rf"
            #print(inst.status)
            inst.remote = False



# -*- coding: utf-8 -*-
"""
    lantz.drivers.prior
    ~~~~~~~~~~~~~~~~~~~

    :company: PRIOR Scientific
    :description: Microscope Automation & Custom Solutions
    :website: http://www.prior.com/

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .nanoscanz import NanoScanZ

__all__ = ['NanoScanZ']



# -*- coding: utf-8 -*-
"""
    lantz.drivers.kentech
    ~~~~~~~~~~~~~~~~~~~~~

    :company: Kentech Instruments Ltd.
    :description: Manufacturers of specialised and custom built electronics and imaging equipment.
    :website: http://www.kentech.co.uk/

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .hri import HRI

__all__ = ['HRI', ]



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.newport
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :company: Newport.
    :description: Test and Measurement Equipment.
    :website: http://www.newport.com/

    ---

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD,

"""

from .powermeter1830c import powermeter1830c

__all__ = ['powermeter1830c']



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.usb
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Implements base classes for drivers that communicate with instruments
    via usb using PyUSB

    See the following link for more information about USB.

    http://www.beyondlogic.org/usbnutshell/usb5.shtml

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from collections import namedtuple, OrderedDict
from fnmatch import fnmatch

import usb
from usb.util import (get_string as usb_get_string,
                      find_descriptor as usb_find_desc)

from lantz import Driver
from lantz.errors import LantzTimeoutError, InstrumentError


ClassCodes = {
    0x00: ('Device', 'Use class information in the Interface Descriptors'),
    0x01: ('Interface', 'Audio'),
    0x02: ('Both', 'Communications and CDC Control'),
    0x03: ('Interface', 'HID (Human Interface Device)'),
    0x05: ('Interface', 'Physical'),
    0x06: ('Interface', 'Image'),
    0x07: ('Interface', 'Printer'),
    0x08: ('Interface', 'Mass Storage'),
    0x09: ('Device', 'Hub'),
    0x0A: ('Interface', 'CDC-Data'),
    0x0B: ('Interface', 'Smart Card'),
    0x0D: ('Interface', 'Content Security'),
    0x0E: ('Interface', 'Video'),
    0x0F: ('Interface', 'Personal Healthcare'),
    0x10: ('Interface', 'Audio/Video Devices'),
    0xDC: ('Both', 'Diagnostic Device'),
    0xE0: ('Interface', 'Wireless Controller'),
    0xEF: ('Both', 'Miscellaneous'),
    0xFE: ('Interface', 'Application Specific'),
    0xFF: ('Both', 'Vendor Specific')
}

# None is 0xxx
AllCodes = {
    (0x00, 0x00, 0x00): 'Use class code info from Interface Descriptors',
    (0x01, None, None): 'Audio device',
    (0x02, None, None): 'Communication device class',
    (0x03, None, None): 'HID device class',
    (0x05, None, None): 'Physical device class',
    (0x06, 0x01, 0x01): 'Still Imaging device',
    (0x07, None, None): 'Printer device',
    (0x08, None, None): 'Mass Storage device',
    (0x09, 0x00, 0x00): 'Full speed Hub',
    (0x09, 0x00, 0x01): 'Hi-speed hub with single TT',
    (0x09, 0x00, 0x02): 'Hi-speed hub with multiple TTs',
    (0x0A, None, None): 'CDC data device',
    (0x0B, None, None): 'Smart Card device',
    (0x0D, 0x00, 0x00): 'Content Security device',
    (0x0E, None, None): 'Video device',
    (0x0F, None, None): 'Personal Healthcare device',
    (0x10, 0x01, 0x00): 'Control Interface',
    (0x10, 0x02, 0x00): 'Data Video Streaming Interface',
    (0x10, 0x03, 0x00): 'VData Audio Streaming Interface',
    (0xDC, 0x01, 0x01): 'USB2 Compliance Device',
    (0xE0, 0x01, 0x01): 'Bluetooth Programming Interface.',
    (0xE0, 0x01, 0x02): 'UWB Radio Control Interface.',
    (0xE0, 0x01, 0x03): 'Remote NDIS',
    (0xE0, 0x01, 0x04): 'Bluetooth AMP Controller.',
    (0xE0, 0x2, 0x01): 'Host Wire Adapter Control/Data interface.',
    (0xE0, 0x2, 0x02): 'Device Wire Adapter Control/Data interface.',
    (0xE0, 0x2, 0x03): 'Device Wire Adapter Isochronous interface.',
    (0xEF, 0x01, 0x01): 'Active Sync device.',
    (0xEF, 0x01, 0x02): 'Palm Sync. This class code can be used in either '
                        'Device or Interface Descriptors.',
    (0xEF, 0x02, 0x01): 'Interface Association Descriptor.',
    (0xEF, 0x02, 0x02): 'Wire Adapter Multifunction Peripheral programming interface.',
    (0xEF, 0x03, 0x01): 'Cable Based Association Framework.',
    (0xEF, 0x04, 0x01): 'RNDIS over Ethernet. Connecting a host to the Internet via '
                        'Ethernet mobile device. The device appears to the host as an'
                        'Ethernet gateway device. This class code may only be used in '
                        'Interface Descriptors.',
    (0xEF, 0x04, 0x02): 'RNDIS over WiFi. Connecting a host to the Internet via WiFi '
                        'enabled mobile device.  The device represents itself to the host'
                        'as an 802.11 compliant network device. This class code may only'
                        'be used in Interface Descriptors.',
    (0xEF, 0x04, 0x03): 'RNDIS over WiMAX. Connecting a host to the Internet via WiMAX '
                        'enabled mobile device.  The device is represented to the host '
                        'as an 802.16 network device. This class code may only be used '
                        'in Interface Descriptors.',
    (0xEF, 0x04, 0x04): 'RNDIS over WWAN. Connecting a host to the Internet via a device '
                        'using mobile broadband, i.e. WWAN (GSM/CDMA). This class code may '
                        'only be used in Interface Descriptors.',
    (0xEF, 0x04, 0x05): 'RNDIS for Raw IPv4. Connecting a host to the Internet using raw '
                        'IPv4 via non-Ethernet mobile device.  Devices that provide raw '
                        'IPv4, not in an Ethernet packet, may use this form to in lieu of '
                        'other stock types. '
                        'This class code may only be used in Interface Descriptors.',
    (0xEF, 0x04, 0x06): 'RNDIS for Raw IPv6. Connecting a host to the Internet using raw '
                        'IPv6 via non-Ethernet mobile device.  Devices that provide raw '
                        'IPv6, not in an Ethernet packet, may use this form to in lieu of '
                        'other stock types. '
                        'This class code may only be used in Interface Descriptors.',
    (0xEF, 0x04, 0x07): 'RNDIS for GPRS. Connecting a host to the Internet over GPRS mobile '
                        'device using the device‚Äôs cellular radio.',
    (0xEF, 0x05, 0x00): 'USB3 Vision Control Interface',
    (0xEF, 0x05, 0x01): 'USB3 Vision Event Interface',
    (0xEF, 0x05, 0x02): 'USB3 Vision Streaming Interface',
    (0xFE, 0x01, 0x01): 'Device Firmware Upgrade.',
    (0xFE, 0x02, 0x00): 'IRDA Bridge device.',
    (0xFE, 0x03, 0x00): 'USB Test and Measurement Device.',
    (0xFE, 0x03, 0x01): 'USB Test and Measurement Device conforming to the USBTMC USB488 Subclass',
    (0xFF, None, None): 'Vendor specific'
}


class LantzUSBTimeoutError(usb.core.USBError, LantzTimeoutError):
    pass


def ep_attributes(ep):
    c = ep.bmAttributes
    attrs = []
    tp = c &usb.ENDPOINT_TYPE_MASK
    if tp == usb.ENDPOINT_TYPE_CONTROL:
        attrs.append('Control')
    elif tp == usb.ENDPOINT_TYPE_ISOCHRONOUS:
        attrs.append('Isochronous')
    elif tp == usb.ENDPOINT_TYPE_BULK:
        attrs.append('Bulk')
    elif tp == usb.ENDPOINT_TYPE_INTERRUPT:
        attrs.append('Interrupt')
        
    sync = (c & 12) >> 2
    if sync == 0:
        attrs.append('No sync')
    elif sync == 1:
        attrs.append('Async')
    elif sync == 2:
        attrs.append('Adaptive')
    elif sync == 3:
        attrs.append('Sync')
    usage = (c & 48) >> 4
    if usage == 0:
        attrs.append('Data endpoint')
    elif usage == 1:
        attrs.append('Feedback endpoint')
    elif usage ==2:
        attrs.append('Subordinate Feedback endpoint')
    elif usage == 3:
        attrs.append('Reserved')

    return ', '.join(attrs)


def list_devices():
    devs = find_devices()
    print('{} devices found'.format(len(devs)))
    for ndx, dev in enumerate(devs):
        print('Device #{}'.format(ndx))
        c, s, p = dev.bDeviceClass, dev.bDeviceSubClass, dev.bDeviceProtocol
        a, b = ClassCodes.get(c, ('Unknown', 'Unknown'))
        nfo = DeviceInfo.from_device(dev)
        print('- Class: {}, {} ({})'.format(a, b, c))
        print('- Subclass and protocol: {} ({} {})'.format(AllCodes.get((c, s, p), 'Unknown'), s, p))
        print('- Manufacturer: {} ({})'.format(nfo.manufacturer, dev.idVendor))
        print('- Product: {} ({})'.format(nfo.product, dev.idProduct))
        print('- Serial Number: {}'.format(nfo.serial_number))
        print('- Number of configurations: {}'.format(dev.bNumConfigurations))
        try:
            for cfg in dev:
                print('- Configurations #{}'.format(cfg.bConfigurationValue))
                for intf in cfg:
                    print('-- Interface: {}'.format(intf.bInterfaceNumber))
                    print('-- Alternate: {}'.format(intf.bAlternateSetting))
                    print('-- Endpoints:')
                    for ep in intf:
                        if usb.util.endpoint_direction(ep.bEndpointAddress) == usb.util.ENDPOINT_OUT:
                            inout = 'out'
                        else:
                            inout = 'in'
                        print('--- {} ({})'.format(ep.bEndpointAddress, inout))
                        print('--- Max Packet Size: {}'.format(ep.wMaxPacketSize))
                        print('--- Polling interval: {}'.format(ep.bInterval))
                        print('--- Attributes: {}'.format(ep_attributes(ep)))
        except Exception as e:
            print('- Exception: {}'.format(e))
        print('-------')


def find_devices(vendor=None, product=None, serial_number=None, custom_match=None, **kwargs):
    """Find connected USB devices matching certain keywords.

    Wildcards can be used for vendor, product and serial_number.

    :param vendor: name or id of the vendor (manufacturer)
    :param product: name or id of the product
    :param serial_number: serial number.
    :param custom_match: callable returning True or False that takes a device as only input.
    :param kwargs: other properties to match. See usb.core.find
    :return:
    """
    kwargs = kwargs or {}
    attrs = {}
    if isinstance(vendor, str):
        attrs['manufacturer'] = vendor
    elif vendor is not None:
        kwargs['idVendor'] = vendor

    if isinstance(product, str):
        attrs['product'] = product
    elif product is not None:
        kwargs['idProduct'] = product

    if serial_number:
        attrs['serial_number'] = str(serial_number)

    if attrs:
        def cm(dev):
            if custom_match is not None and not custom_match(dev):
                return False
            info = DeviceInfo.from_device(dev)
            for attr, pattern in attrs.items():
                if not fnmatch(getattr(info, attr).lower(), pattern.lower()):
                    return False
            return True
    else:
        cm = custom_match

    return usb.core.find(find_all=True, custom_match=cm, **kwargs)


def find_interfaces(device, **kwargs):
    """
    :param device:
    :return:
    """
    interfaces = []
    try:
        for cfg in device:
            try:
                interfaces.extend(usb_find_desc(cfg, find_all=True, **kwargs))
            except:
                pass
    except:
        pass
    return interfaces


def find_endpoint(interface, direction, type):
    ep = usb_find_desc(interface, custom_match=
                                  lambda e: usb.util.endpoint_direction(e.bEndpointAddress) == direction and
                                            usb.util.endpoint_type(e.bmAttributes) == type
         )
    return ep


class DeviceInfo(namedtuple('DeviceInfo', 'manufacturer product serial_number')):

    def __str__(self):
        return '{} {}. S/N {}'.format(self.manufacturer, self.product, self.serial_number)

    @classmethod
    def from_device(cls, device):
        ret = []
        for attr in ('iManufacturer', 'iProduct', 'iSerialNumber'):
            try:
                ret.append(usb_get_string(device, 255, getattr(device, attr)).strip())
            except Exception as e:
                ret.append('Unknown')
        return cls(*ret)


class USBDriver(Driver):
    """Base class for drivers that communicate with instruments
    via usb port using pyUSB
    """

    #: Configuration number to be used. If None, the default will be used.
    CONFIGURATION = None
    #: Interface index it be used
    INTERFACE = (0, 0)
    #: Receive and Send endpoints to be used. If None the first IN (or OUT) BULK
    #: endpoint will be used.
    ENDPOINTS = (None, None)

    TIMEOUT = 1000

    find_devices = staticmethod(find_devices)

    def __init__(self, vendor=None, product=None, serial_number=None,
                 device_filters=None, timeout=None, **kwargs):
        super().__init__(**kwargs)

        device_filters = device_filters or {}
        devices = self.find_devices(vendor, product, serial_number, None, **device_filters)

        if not devices:
            raise errors.InstrumentError('No device found.')
        elif len(devices) > 1:
            desc = '\n'.join(str(DeviceInfo.from_device(dev)) for dev in devices)
            raise errors.InstrumentError('{} devices found:\n{}\n'
                                  'Please narrow the search criteria'.format(len(devices), desc))

        self.usb_dev, other = devices[0], devices[1:]
        nfo = DeviceInfo.from_device(self.usb_dev)
        self.log_debug('- Manufacturer: {} ({})'.format(nfo.manufacturer, self.usb_dev.idVendor))
        self.log_debug('- Product: {} ({})'.format(nfo.product, self.usb_dev.idProduct))
        self.log_debug('- Serial Number: {}'.format(nfo.serial_number))
        try:
            if self.usb_dev.is_kernel_driver_active(0):
                    self.usb_dev.detach_kernel_driver(0)
        except (usb.core.USBError, NotImplementedError) as e:
            self.log_warning(repr(e))

        try:
            self.usb_dev.set_configuration() #self.CONFIGURATION
            self.usb_dev.set_interface_altsetting()
        except usb.core.USBError as e:
            self.log_error("Could not set configuration")
            raise errors.InstrumentError('failed to set configuration')

        self.usb_intf = self._find_interface(self.usb_dev, self.INTERFACE)
        self.log_debug('Interface: {}'.format(self.usb_intf.index))


        self.usb_recv_ep, self.usb_send_ep = self._find_endpoints(self.usb_intf, self.ENDPOINTS)

        self.log_debug('EP Address: recv={}, send={}'.format(self.usb_recv_ep.bEndpointAddress,
                                                             self.usb_send_ep.bEndpointAddress))

    def _find_interface(self, dev, setting):
        return self.usb_dev.get_active_configuration()[self.INTERFACE]

    def _find_endpoints(self, interface, setting):
        recv, send = setting
        if recv is None:
            recv = find_endpoint(interface, usb.ENDPOINT_IN, usb.ENDPOINT_TYPE_BULK)
        else:
            recv = usb_find_desc(interface, bEndpointAddress=recv)

        if send is None:
            send = find_endpoint(interface, usb.ENDPOINT_OUT, usb.ENDPOINT_TYPE_BULK)
        else:
            send = usb_find_desc(interface, bEndpointAddress=send)

        return recv, send

    def send(self, data):
        """Send raw bytes to the instrument.

        :param data: bytes to be sent to the instrument
        :type data: bytes
        """
        self.raw_send(data)

    def raw_send(self, data):
        """Send raw bytes to the instrument.

        :param data: bytes to be sent to the instrument
        :type data: bytes
        """

        try:
            return self.usb_send_ep.write(data)
        except usb.core.USBError as e:
            raise errors.InstrumentError(str(e))

    def recv(self, size):
        return self.raw_recv(size)

    def raw_recv(self, size):
        """Receive raw bytes to the instrument.

        :param size: number of bytes to receive
        :return: received bytes
        :return type: bytes
        """

        if size <= 0:
            size = 1

        data = self.usb_recv_ep.read(size, self.TIMEOUT).tobytes()

        return data

    def finalize(self):
        """Close port
        """
        self.log_debug('Closing device {}', str(DeviceInfo.from_device(self.usb_dev)))
        return usb.util.dispose_resources(self.usb_dev)


def _patch_endpoint(ep, log_func=print):
    _read = ep.read
    _write = ep.write
    def new_read(*args, **kwargs):
        log_func('---')
        log_func('reading from {}'.format(ep.bEndpointAddress))
        log_func('args: {}'.format(args))
        log_func('kwargs: {}'.format(kwargs))
        ret =_read(*args, **kwargs)
        log_func('returned', ret)
        log_func('---')
        return ret
    def new_write(*args, **kwargs):
        log_func('---')
        log_func('writing to {}'.format(ep.bEndpointAddress))
        log_func('args: {}'.format(args))
        log_func('kwargs: {}'.format(kwargs))
        ret = _write(*args, **kwargs)
        log_func('returned', ret)
        log_func('---')
        return ret
    ep.read = new_read
    ep.write = new_write



"""
    lantz.drivers.ni.daqmx.utils
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    These are simple implementation of commons functionalities.

    Author: Alexandre Bourassa
    Date: 04/08/2017
"""
from lantz import Driver
from lantz.driver import Action
from lantz.drivers.ni.daqmx import DigitalOutputTask, DigitalOutputChannel
import numpy as np

class DigitalSwitch(Driver):

    def __init__(self, ch='/dev1/po.0'):
        super().__init__()
        self.task = DigitalOutputTask()
        self.task.add_channel(DigitalOutputChannel(ch))
        clock_config = {
            'source': 'OnboardClock',
            'rate': 10000,
            'sample_mode': 'finite',
            'samples_per_channel': 100,
        }

    @Action()
    def output(self, state):
    	if state:
    		state_pts = np.ones(100)
    	else:
    		state_pts = np.zeros(100)
        with self.task as task:
            self.task.write(state_pts)


# -*- coding: utf-8 -*-
"""
    lantz.drivers.newport.motion axis
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    General class that implements the commands used for motion
    drivers

    :copyright: 2018, see AUTHORS for more details.
    :license: GPL, see LICENSE for more details.


"""


from lantz.feat import Feat
from lantz.action import Action
from lantz.driver import Driver
from lantz import Q_
from lantz.processors import convert_to
import time
import numpy as np

#  Add generic units:
# ureg.define('unit = unit')
# ureg.define('encodercount = encodercount')
# ureg.define('motorstep = motorstep')


class MotionAxisSingle(Driver):
    def __init__(self, *args, **kwargs):
        self.wait_time = 0.01  # in seconds * Q_(1, 's')
        self.wait_until_done = True
        self.accuracy = 0.001  # in units reported by axis
        self._units = 'mm'

    @Feat()
    def idn(self):
        return self.query('ID?')

    @Feat()
    def position(self):
        raise NotImplementedError

    @position.setter
    def position(self, pos):
        """
        Waits until movement is done if self.wait_until_done = True.

        :param pos: new position
        """
        self._set_position(pos, wait=self.wait_until_done)

    @Action(units=['mm', None])
    def _set_position(self, pos, wait=None):
        """
        Move to an absolute position, taking into account backlash.

        When self.backlash is to a negative value the stage will always move
         from low to high values. If necessary, a extra step with length
         self.backlash is set.

        :param pos: New position in mm
        :param wait: wait until stage is finished
        """

        # First do move to extra position if necessary

        self.__set_position(pos)
        if wait:
            self._wait_until_done()
            self.check_position(pos)

    def __set_position(self, pos):
        """
        Move stage to a certain position
        :param pos: New position
        """
        self.write('PA%f' % (pos))

    @Action(units='mm')
    def check_position(self, pos):
        '''Check is stage is at expected position'''
        if np.isclose(self.position, pos, atol=self.accuracy):
            return True
        self.log_error('Position accuracy {} is not reached.'
                       'Expected: {}, measured: {}'.format(self.accuracy,
                                                           pos,
                                                           self.position))
        return False


    @Feat(values={True: '1', False: '0'})
    def motion_done(self):
        raise NotImplementedError

    @Feat()
    def units(self):
        return self._units

    @units.setter
    def units(self, units):
        super().update_units(self._units, units)
        self._units = units

    def _wait_until_done(self):
        wait_time = convert_to('seconds', on_dimensionless='ignore')(self.wait_time)
        time.sleep(wait_time)
        while not self.motion_done:
            time.sleep(wait_time)
        return True


class MotionAxisMultiple(MotionAxisSingle):
    def __init__(self, parent, num, id, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent = parent
        self.num = num
        self._idn = id

    def __del__(self):
        self.parent = None
        self.num = None

    def query(self, command, *, send_args=(None, None), recv_args=(None, None)):
        return self.parent.query('{:d}{}'.format(self.num, command),
                                 send_args=send_args, recv_args=recv_args)

    def write(self, command, *args, **kwargs):
        return self.parent.write('{:d}{}'.format(self.num, command),
                                 *args, **kwargs)

class BacklashMixing():
    '''Adds functionality to a motionaxis: blacklash correction'''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.backlash = 0

    @Action(units=['mm', None])
    def _set_position(self, pos, wait=None):
        """
        Move to an absolute position, taking into account backlash.

        When self.backlash is to a negative value the stage will always move
         from low to high values. If necessary, a extra step with length
         self.backlash is set.

        :param pos: New position in mm
        :param wait: wait until stage is finished
        """

        # First do move to extra position if necessary
        if self.backlash:
            position = self.position.magnitude
            backlash = convert_to(self.units, on_dimensionless='ignore'
                                   )(self.backlash).magnitude
            if (backlash < 0 and position > pos) or\
               (backlash > 0 and position < pos):

                self.log_info('Using backlash')
                self.__set_position(pos + backlash)
                self._wait_until_done()

        # Than move to final position
        self.__set_position(pos)
        if wait:
            self._wait_until_done()
            self.check_position(pos)

    def __set_position(self, pos):
        """
        Move stage to a certain position
        :param pos: New position
        """
        self.write('PA%f' % (pos))
        self.last_set_position = pos


from lantz import Driver, Feat, DictFeat, Action

import ctypes as ct

from ctypes import c_char_p, c_void_p, c_int, c_double, POINTER, byref, create_string_buffer

from collections import Iterable, OrderedDict


class PhotonEtcComm():


    def __init__(self):

        # this should point to the location of the DLL
        dll_loc = 'D:\\Code\\Github\\lantz\\lantz\\drivers\\photonetc\\PE_Filter_SDK.dll'
        self.config_dir = 'C:\\Program Files (x86)\\Photon etc\\PHySpecV2\\Devices\\'


        # Uses WinDLL to load the dll file
        self.lib = ct.WinDLL(dll_loc)

        # Here, we set up the argument types for the varios functions
        # encapsulated in the DLL. All these are essentially taken from the
        # header file PE_Filter.h

        # Functions necessary for setting up PE filter class
        self.lib.PE_Create.argtypes = [c_char_p, c_void_p]
        self.lib.PE_Destroy.argtypes = [c_void_p]

        self.lib.PE_GetSystemCount.argtypes = [c_void_p]
        self.lib.PE_GetSystemName.argtypes = [c_void_p, c_int, c_char_p, c_int]

        self.lib.PE_GetStatusStr.argtypes = [c_int]

        # These filters are used for interacting with a single filter moodule.
        self.lib.PE_Open.argtypes = [c_void_p, c_char_p]
        self.lib.PE_Close.argtypes = [c_void_p]

        self.lib.PE_GetWavelength.argtypes = [c_void_p, POINTER(c_double)]
        self.lib.PE_SetWavelength.argtypes = [c_void_p, c_double]

        # These settings enable control of ome of the more advanced settings/
        # options, and the different gratings present in the filter.
        self.lib.PE_GetWavelengthRange.argtypes = [c_void_p, POINTER(c_double), POINTER(c_double)]
        self.lib.PE_HasHarmonicFilter.argtypes = [c_void_p]

        self.lib.PE_GetHarmonicFilterEnabled.argtypes = [c_void_p, POINTER(c_int)]
        self.lib.PE_SetHarmonicFilterEnabled.argtypes = [c_void_p, c_int]

        self.lib.PE_GetGratingCount.argtypes = [c_void_p, POINTER(c_int)]
        self.lib.PE_GetGratingCount.argtypes = [c_void_p, c_int, c_char_p, c_int]
        self.lib.PE_GetGratingWavelengthRange.argtypes = [c_void_p, c_int, POINTER(c_double), POINTER(c_double)]
        self.lib.PE_GetGratingWavelengthExtendedRange.argtypes = [c_void_p, c_int, POINTER(c_double), POINTER(c_double)]

        self.lib.PE_SetWavelengthOnGrating.argtypes = [c_void_p, c_int, c_double]
        self.lib.PE_GetGrating.argtypes = [c_void_p, POINTER(c_int)]



    def __del__(self):
        """
        Calls C garbage collection methods from Python in order to close the LLTF
        communication module, and avoid issues with recollecting.
        """
        #self.destroy()

    def getLibraryVersion(self):
        """
        Returns the version of the library used by the software.
        """
        value =  hex(self.lib.PE_GetLibraryVersion())

        # code to decode bit shifted string into version
        major = int(value[2:-4], 16)
        minor = int(value[-4:-2], 16)
        bugfix = int(value[-2:], 16)
        return '{}.{}.{}'.format(major, minor, bugfix)


    def create(self, system):
        """
        Creates a handle for the filter described in "system".xml configuration file.
        """

        xml_filename = self.config_dir + system + '.xml'

        # create pointer to serve as system handle
        pe_handle = c_void_p()

        # encode configuration filename, pass reference to system handle
        status = self.lib.PE_Create(xml_filename.encode(), byref(pe_handle))

        return pe_handle, status

    def destroy(self, pe_handle):
        """
        Destroys the filter handle, for garbage collection purposes.
        """
        status = self.lib.PE_Destroy(pe_handle)
        return status


    def getSystemName(self, pe_handle, index=0):
        """

        """
        size = 256
        systemname = create_string_buffer(size)

        status = self.lib.PE_GetSystemName(pe_handle, index, systemname, size)

        return systemname.value.decode('utf-8'), status


    def openSystem(self, pe_handle, systemname):

        status = self.lib.PE_Open(pe_handle, systemname.encode())
        return status

    def closeSystem(self, pe_handle):

        status = self.lib.PE_Close(pe_handle)
        return status


    def getWavelength(self, pe_handle):

        wavelength = c_double()

        status = self.lib.PE_GetWavelength(pe_handle, byref(wavelength))

        return wavelength.value, status

    def setWavelength(self, pe_handle, wavelength):

        status = self.lib.PE_SetWavelength(pe_handle, wavelength)

        return status


    def getWavelengthRange(self, pe_handle):

        min_wavelength = c_double()
        max_wavelength = c_double()

        status = self.lib.PE_GetWavelengthRange(pe_handle, byref(min_wavelength), byref(max_wavelength))

        return (min_wavelength.value, max_wavelength.value), status



    def statusCode(self, code):
        """
        Implements lookup table for status codes. Note that these must be kept
        up to date with any changes that may occur in the C header file, since
        they are currently hard coded here and do not draw from PE_filter.h!
        """

        # this doesn't seem to work - not sure why, but appears to return garbage.
        #self.lib.PE_GetStatusStr(code)

        STATUS_CODES = {
            0: 'Successful operation',
            1: 'Invalid handle: handle already deleted or null',
            2: 'PE Failure: instrument communication failure',
            3: 'Configuration file is missing.',
            4: 'Configuration file invalid.',
            5: 'Invalid wavelength.',
            6: 'Missing harmonic filter: no harmonic filter present in system.',
            7: 'Invalid filter: requested filter does not exist.',
            8: 'Unknown status',
            9: 'Invalid grating: requested grating does not exist',
            10: 'Invalid buffer: null buffer',
            11: 'Unsupported filter configuration.',
            12: 'Filter connection error: no filter is connected.'
        }
        #status_str = self.lib.PE_GetStatusStr(code)

        return STATUS_CODES[code]



class PhotonEtcFilter(Driver):

    def initialize(self):
        """
        Creates the library for communicating with the tunable filter.
        """
        self.comm = PhotonEtcComm()

        self.filters = {}

    def add_filter(self, system_name, index=0, user_name=None):
        """
        Connects to the filter at system_name, index to the system.
        """

        if not user_name:

            user_name = system_name

        # create PE_Filter
        pe_handle, status = self.comm.create(system_name)

        # get filter name at index
        name, status = self.comm.getSystemName(pe_handle, index)

        # open connection to the system
        status = self.comm.openSystem(pe_handle, name)

        self.filter_info = {'name': system_name,
                            'wavelength_range': self.comm.getWavelengthRange(pe_handle)[0],
                            'user_name': user_name,
                            'handle': pe_handle}

        self.filters[user_name] = self.filter_info

        print(self.filter_info)

    def remove_filter(self, filter_name):
        """
        Removes the specified filter to avoid memory leak issues.
        """
        status = self.comm.closeSystem(filter_name['handle'])

        if len(list(self.filters.keys())) == 0:
            status = self.comm.destroy(filter_name['handle'])

        return status

    def finalize(self):
        """
        Closes the connections to PE filters and the PE filter library.
        """
        for key in list(self.filters.keys()):

            f = self.filters.pop(key)
            print('Closing time!')
            self.remove_filter(f)

    @Feat(read_once=True)
    def lib_verision(self):
        """
        Test code to check library version.
        """
        return self.comm.getLibraryVersion()

    @DictFeat(units='nm')#, keys=FILTERS)
    def wavelength(self, filter_name):
        """
        Returns the filter wavelength in nanometers. Prints an error message
        if the returned value lies outside the filter range.
        """
        handle = self.filters[filter_name]['handle']

        wavelength, status = self.comm.getWavelength(handle)

        min_nm, max_nm = self.wavelength_range['IR']

        if min_nm <= wavelength <= max_nm:

            return wavelength

        print('Invalid wavelength {}nm for {}, must lie between {}nm and {}nm'.format(wavelength, filter_name, min_nm, max_nm))

    @wavelength.setter
    def wavelength(self, filter_name, nm):
        """
        Sets the filter wavelength to be nm (in nanometers). Includes bounds
        checking to ensure that the set wavelength falls within the range
        supported by the filter.
        """

        handle = self.filters[filter_name]['handle']

        min_nm, max_nm = self.wavelength_range[filter_name]

        if min_nm <= nm <= max_nm:

            return self.comm.setWavelength(handle, nm)

        print('Invalid wavelength {}nm for {}, must lie between {}nm and {}nm'.format(nm, filter_name, min_nm, max_nm))


    @DictFeat(read_once = True)
    def wavelength_range(self, filter_name):
        """
        Returns the wavelength range of the specified filter.
        """
        handle = self.filters[filter_name]['handle']

        wavelength_range, status = self.comm.getWavelengthRange(handle)
        return wavelength_range



def test():
    # test script tries to implement example code found in SDK manual as proof
    # of concept through lantz interface w/ DictFeat structure to handle
    # multiple filters at a single time.

    ## TODO: test code and see if it can support simultaneous control of both
    # filters without issues.

    lltf = PhotonEtcFilter()
    lltf.initialize()

    print('Library version: {}'.format(lltf.lib_verision))

    system_list = ['M000010133']  #,'M000010161']
    filter_name = 'IR'

    for f in system_list:

        lltf.add_filter(f, user_name=filter_name)

        print('Wavelength range: {}'.format(lltf.wavelength_range[filter_name]))

        print('Wavelength: {}'.format(lltf.wavelength[filter_name]))

        lltf.wavelength[filter_name] = 850.0

        print('Wavelength: {}'.format(lltf.wavelength[filter_name]))
        lltf.wavelength[filter_name] = 950.0

        print('Wavelength: {}'.format(lltf.wavelength[filter_name]))

        lltf.wavelength[filter_name] = 2400.0

        print('Wavelength: {}'.format(lltf.wavelength[filter_name]))

        # this should print an error message
        lltf.wavelength[filter_name] = 400.0

    lltf.finalize()




if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
"""
    lantz.drivers.example.serial_example
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Serial example.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz import Action, Feat, DictFeat
from lantz import MessageBasedDriver


class SerialTemplate(MessageBasedDriver):
    """Template for drivers connecting via serial port.
    """


    DEFAULTS = {'ASRL': {'write_termination': '\n',
                                'read_termination': '\n'}}

    @Feat()
    def a_read_only_property(self):
        """Help for a a_read_only_property
        """
        return self.query('*IDN?')

    @Feat(units='V', limits=(10,))
    def a_read_write_property(self):
        """Help for a_read_write_property
        """
        return float(self.query('?AMP'))

    @a_read_write_property.setter
    def a_read_write_property(self, value):
        self.query('!AMP {:.1f}'.format(value))

    @DictFeat(values={True: '1', False: '0'}, keys=list(range(1,9)))
    def a_read_write_dictionary_property(self, key):
        """Help for a_read_write_dictionary_property
        """
        return self.query('?DOU {}'.format(key))

    @a_read_write_dictionary_property.setter
    def a_read_write_dictionary_property(self, key, value):
        self.query('!DOU {} {}'.format(key, value))

    @Action()
    def do_something(self):
        """Help for do_something
        """
        return


if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test Kentech HRI')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False, help='Show interactive GUI')
    parser.add_argument('-p', '--port', type=str, default='17',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantz.log.log_to_socket(lantz.log.DEBUG)
    with SerialTemplate(args.port) as inst:
        if args.interactive:
            from lantz.ui.app import start_test_app
            start_test_app(inst)
        else:
            # Add your test code here
            print('Non interactive mode')





# -*- coding: utf-8 -*-
"""
    lantz.drivers.examples
    ~~~~~~~~~~~~~~~~~~~~~~

    :company: Lantz Examples.
    :description: Example drivers for simulated instruments.
    :website:

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .fungen import LantzSignalGenerator
from .voltmeter import LantzVoltmeter

__all__ = ['LantzSignalGenerator', 'LantzVoltmeter']



from lantz.driver import Driver
from lantz import Feat, Action
import cv2
import numpy as np

class USBCam(Driver):

    def __init__(self, device_id):
        self.device_id = device_id
        self._flipud = False
        self._fliplr = False
        self._rotation = 0
        return

    def initialize(self):
        self.capture = cv2.VideoCapture(self.device_id)
        return

    def finalize(self):
        self.capture.release()
        return

    @Feat(values={0, 90, 180, 270})
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, value):
        self._rotation = value
        return

    @Feat(values={True, False})
    def flipud(self):
        return self._flipud

    @flipud.setter
    def flipud(self, value):
        self._flipud = value
        return

    @Feat(values={True, False})
    def fliplr(self):
        return self._fliplr

    @fliplr.setter
    def fliplr(self, value):
        self._fliplr = value
        return

    @Action()
    def get_frame(self):
        img = self.capture.read()[1]
        array = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        if self._flipud:
            array = np.flipud(array)
        if self._fliplr:
            array = np.fliplr(array)
        array = np.rot90(array, k=int(self._rotation / 90))
        return array



import winspec as ws
import socket
import numpy as np
import pickle
import json
import time

print "Winspec is imported"
socket.setdefaulttimeout(None)

def initiate_server():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    host = socket.gethostname()
    IP = socket.gethostbyname(host)
    port = 12345
    s.settimeout(300)
    print "The IP address is ", IP
    print "The SOCKET number is ", port
    print "Waiting for connection ..."
    s.bind((host, port))
    s.listen(5)
    c, addr = s.accept()
    print "Connection Established with ", addr
    s.settimeout(None)
    return c


no_arg_functions = {
    "get exposure time": ws.get_exposure_time,
    "get grating": ws.get_grating,
    "get ngrating": ws.get_ngratings,
    "get current turret": ws.get_current_turret,
    "get temperature": ws.get_temperature,
    "get target temperature": ws.get_target_temperature,
    "get wavelength": ws.get_wavelength
    }

one_arg_send_functions = {
    "set exposure time": ws.set_exposure_time,
    "set grating": ws.set_grating,
    "set target temperature": ws.set_target_temperature,
    "set wavelength": ws.set_wavelength
    }

one_arg_get_functions = {
    "get grating grooves": ws.get_grating_grooves,
    "get grating name": ws.get_grating_name
    }

def get_spectrum():
    spectrum = ws.get_spectrum()
    return json.dumps(spectrum.tolist())

def get_wavelengths():
    wavelengths = ws.get_wavelengths()
    return json.dumps(wavelengths)

def one_arg_function(command):
    split_list = command.split()
    val = float(split_list[-1])
    function = " ".join(split_list[0:-1])
    res = one_arg_functions[function](val)
    return str(res)

def parse_one_arg_functions(command):
    split_list = command.split()
    val = split_list[-1]
    function = " ".join(split_list[0:-1])
    return function, val    

def no_arg_function(command):
    return str(no_arg_functions[command]())

def recieve_and_send(command):
    if command == "get spectrum":
        return get_spectrum()
    elif command == "get wavelengths":
        return get_wavelengths()
    else:
        if command in no_arg_functions.keys():
            return no_arg_function(command)
        else:
            command, val = parse_one_arg_functions(command)
            if command in one_arg_send_functions.keys():
                one_arg_send_functions[command](float(val))
                return "OK"
            elif command in one_arg_get_functions.keys():
                return str(one_arg_get_functions[command](int(val)))
            else:
                return "Unidennd"

def establish_communication():
    c = initiate_server()
    try:
        socketerror1035count = 0
        while True:
            try:
                command = c.recv(1024)
                #print command
                if not command: break
                if command == "close":
                    break
                to_be_sent = recieve_and_send(command) +"\r\n"
                #print to_be_sent
                c.send(to_be_sent)
            except socket.error as e:
                if e[0] == 10035 and socketerror1035count < 5:
                    socketerror1035count = socketerror1035count  +1
                    time.sleep(1)
    finally:
        c.close()

if __name__ == "__main__":
    while True:
        establish_communication()
        time.sleep(0.5)



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.aa.aotf
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the drivers for an AOTF Controller


    Implementation Notes
    --------------------

    There are currently two (disconnected) ways of setting the power for each
    line: powerdb and power.

    Sources::

        - MDSnC Manual

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

#TODO: Implement calibrated power.

from lantz import Feat, DictFeat
from lantz.drivers.legacy.serial import SerialDriver

class MDSnC(SerialDriver):
    """MDSnC synthesizer for AOTF.nC
    """

    CHANNELS = list(range(8))

    @Feat(None, values={True: 1, False: 0})
    def main_enabled(self, value):
        """Enable the
        """
        self.send("I{}".format(value))

    @DictFeat(None, keys=CHANNELS)
    def enabled(self, channel, value):
        """Enable single channels.
        """
        self.send("L{}O{}".format(channel, value))

    @DictFeat(None, keys=CHANNELS)
    def frequency(self, channel, value):
        """RF frequency for a given channel.
        """
        self.send("L{}F{}".format(channel, value))

    @DictFeat(None, keys=CHANNELS)
    def powerdb(self, channel, value):
        """Power for a given channel (in db).
        """
        self.send("L{}D{}".format(channel, value))

    @DictFeat(None, keys=CHANNELS, limits=(0, 1023, 1))
    def power(self, channel, value):
        """Power for a given channel (in digital units).
        """
        self.send("L{}P{:04d}".format(channel, value))


if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test Kentech HRI')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False, help='Show interactive GUI')
    parser.add_argument('-p', '--port', type=str, default='17',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantz.log.log_to_socket(lantz.log.DEBUG)
    with MDSnC(args.port) as inst:
        if args.interactive:
            from lantz.ui.app import start_test_app
            start_test_app(inst)
        else:
            from time import sleep
            print("init")
            freq = 130
            inst.power(4,10)
            sleep(0.2)
            inst.enabled(4,1)
            sleep(1.2)
            inst.enabled(4,0)
            sleep(1.2)
            inst.enabled(4,1)





# -*- coding: utf-8 -*-
"""
    lantz.drivers.photonetc
    ~~~~~~~~~~~~~~~~~~~~~~
    :company: Photon Etc.
    :description: Manufacturer of infrared cameras, hyperspectral imaging,
    and spectroscopic  scientific instruments.
    :website: http://www.photonetc.com/
    ----
    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from .lltf import PhotonEtcComm
from .lltf import PhotonEtcFilter

__all__ = ['PhotonEtcComm', 'PhotonEtcFilter']



# -*- coding: utf-8 -*-
"""
    lantz.drivers.andor
    ~~~~~~~~~~~~~~~~~~~

    :company: Andor
    :description: Scientific cameras.
    :website: http://www.andor.com/

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .andor import Andor
from .neo import Neo
from .ccd import CCD

__all__ = ['Andor', 'Neo', 'CCD']




# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.visa
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements base classes for drivers that communicate with instruments using visalib.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz import Driver
from lantz.drivers.legacy.textual import TextualMixin
from lantz.errors import LantzTimeoutError

import visa


class LantzVisaTimeoutError(LantzTimeoutError):
    pass

"""
BYTESIZE = {5: 5, 6: 6,
            7: 7, 8: 8}

PARITY = {'none': Constants.ASRL_PAR_NONE, 'even': Constants.ASRL_PAR_EVEN,
          'odd': Constants.ASRL_PAR_ODD, 'mark': Constants.ASRL_PAR_MARK,
          'space': Constants.ASRL_PAR_SPACE}

STOPBITS = {1: Constants.ASRL_STOP_ONE, 1.5: Constants.ASRL_STOP_ONE5,
            2: Constants.ASRL_STOP_TWO}
"""

class VisaDriver(object):

    def __new__(cls, resource_name, *args, **kwargs):
        library_path = kwargs.get('library_path', None)
        if library_path:
            manager = visa.ResourceManager(library_path)
        else:
            manager = visa.ResourceManager()
        name = manager.resource_info(resource_name).resource_name
        if name.startswith('GPIB'):
            return GPIBVisaDriver(resource_name, *args, **kwargs)
        elif name.startswith('ASRL'):
            return SerialVisaDriver(resource_name, *args, **kwargs)
        elif name.startswith('TCPIP'):
            return TCPVisaDriver(resource_name, *args, **kwargs)
        elif name.startswith('USB'):
            return USBVisaDriver(resource_name, *args, **kwargs)
        else:
            raise ValueError('Unknown resource type: {}'.format(name))


class MessageVisaDriver(TextualMixin, Driver):
    """Base class for drivers that communicate with instruments
    via serial or parallel port using pyserial

    :param resource_name: name or alias of the resource to open.

    """

    RECV_TERMINATION = '\n'
    SEND_TERMINATION = '\n'
    ENCODING = 'ascii'

    RECV_CHUNK = -1

    def __init__(self, resource_name, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._init_attributes = {}

        library_path = kwargs.get('library_path', None)
        if library_path:
            self.resource_manager = visa.ResourceManager(library_path)
        else:
            self.resource_manager = visa.ResourceManager()

        self.resource = None

        self.resource_name = resource_name
        self.log_debug('Created Instrument {}', self.resource_name)

    def raw_send(self, data):
        """Send raw bytes to the instrument.

        :param data: bytes to be sent to the instrument
        :param data: bytes
        """

        try:
            self.resource.write_raw(data)
        except Exception as e:
            raise Exception(str(e))

    def initialize(self):
        """Open port
        """
        if not self.is_open():
            self.log_debug('Opening {}', self.resource_name)
            self.resource = self.resource_manager.open_resource(self.resource_name)
            for key, value in self._init_attributes.items():
                self.resource.set_visa_attribute(key, value)

            self.log_debug('The session for {} is {}', self.resource_name, self.resource.session)
        else:
            self.log_debug('{} is already open', self.resource_name)

    def finalize(self):
        """Close port
        """
        self.log_debug('Closing port {}', self.resource_name)
        self.resource.close()

    def is_open(self):
        if self.resource is None:
            return False
        return self.resource.session is not None


class SerialVisaDriver(MessageVisaDriver):
    """Base class for drivers that communicate with instruments
    via serial port using visa.

    :param resource_name: the visa resource name or alias (e.g. 'ASRL1::INSTR')
    """

    #: communication parameters
    BAUDRATE = 9600
    BYTESIZE = 8
    PARITY = 'none'
    STOPBITS = 1

    #: flow control flags
    RTSCTS = False
    DSRDTR = False
    XONXOFF = False

    RECV_CHUNK = -1

    def __init__(self, resource_name, *args, **kwargs):
        super().__init__(resource_name, *args, **kwargs)

        kw = {}
        kw['ASRL_BAUD']= kwargs.get('baudrate', self.BAUDRATE)
        kw['ASRL_DATA_BITS'] = BYTESIZE[kw.get('bytesize', self.BYTESIZE)]
        kw['ASRL_PARITY'] = PARITY[kw.get('parity', self.PARITY)]
        kw['ASRL_STOP_BITS'] = STOPBITS[kw.get('stopbits', self.STOPBITS)]

        flow = Constants.ASRL_FLOW_NONE
        if kwargs.get('rtscts', getattr(self, 'RTSCTS')):
            flow |= Constants.ASRL_FLOW_RTS_CTS
        if kwargs.get('dsrdtr', getattr(self, 'DSRDTR')):
            flow |= Constants.ASRL_FLOW_DTR_DSR
        if kwargs.get('xonxoff', getattr(self, 'XONXOFF')):
            flow |= Constants.ASRL_FLOW_XON_XOFF

        kw['ASRL_FLOW_CNTRL'] = flow

        if self.RECV_TERMINATION and self.RECV_CHUNK > 1:
            kw['TERMCHAR'] = ord(self.RECV_TERMINATION)
            kw['ASRL_END_IN'] = Constants.ASRL_END_TERMCHAR
        else:
            kw['ASRL_END_IN'] = Constants.ASRL_END_NONE

        self._init_attributes.update(kw)


    def raw_recv(self, size):
        """Receive raw bytes to the instrument.

        :param size: number of bytes to receive
        :return: received bytes
        :return type: bytes

        If a timeout is set, it may return less bytes than requested.
        If size == -1, then the number of available bytes will be read.

        """

        if size == -1:
            size = self.visa.get_attribute(self.vi, 'ASRL_AVAIL_NUM')
            if not size:
                return bytes()

        if not size:
            size = 1

        data = self.visa.read(self.vi, size)

        return data


class GPIBVisaDriver(MessageVisaDriver):


    def raw_recv(self, size):
        """Receive raw bytes to the instrument.

        :param size: number of bytes to receive
        :return: received bytes
        :return type: bytes

        If a timeout is set, it may return less bytes than requested.
        If size == -1, then the number of available bytes will be read.

        """

        if not size:
            size = 1

        data = self.resource.read_raw(1)

        return data

class TCPVisaDriver(MessageVisaDriver):
    pass


class USBVisaDriver(MessageVisaDriver):


    def raw_recv(self, size):
        """Receive raw bytes to the instrument.

        :param size: number of bytes to receive
        :return: received bytes
        :return type: bytes

        If a timeout is set, it may return less bytes than requested.
        If size == -1, then the number of available bytes will be read.

        """

        if not size:
            size = 1

        data = self.resource.read_raw(1)

        return data



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.tektronix.tds1012
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the drivers to control an oscilloscope.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
    
    Source: Tektronix Manual
"""

import numpy as np

from lantz.feat import Feat
from lantz.action import Action
from lantz.drivers.legacy.serial import SerialDriver
from lantz.errors import InvalidCommand

class TDS1012(SerialDriver):
    """Tektronix TDS1012 100MHz 2 Channel Digital Storage Oscilloscope
    """
    ENCODING = 'ascii'

    RECV_TERMINATION = '\n'
    SEND_TERMINATION = '\n'
    TIMEOUT = -1    # Avoids timeout while acquiring a curve. May not be the 
                    # best option.
    
    def __init__(self, port):
       # super().TIMEOUT = 20
        super().__init__(port)
        super().initialize() # Automatically open the port
        
    @Action()
    def initiate(self):
        """ Initiates the acquisition in the osciloscope.
        """
        self.send(':ACQ:STATE ON')
            
    @Action()
    def idn(self):
        """ Identify the Osciloscope
        """
        return self.query('*IDN?')
        
    @Action()
    def autoset(self):
        """ Adjust the vertical, horizontal and trigger controls to display a 
            stable waveform.
        """
        self.send('AUTOS EXEC')
    
    @Action()
    def autocal(self):
        """ Autocalibration of osciloscope. It may take several minutes to 
            complete
        """
        return self.send('*CAL')
       
    @Feat(limits=(1,2))
    def datasource(self):
        """ Retrieves the data source from which data is going to be taken. 
            TDS1012 has 2 channels
        """
        return self.query('DAT:SOU?')
    
    @datasource.setter
    def datasource(self,value):
        """ Sets the data source for the acquisition of data.
        """
        self.send('DAT:SOU CH{}'.format(value))
    
    @Action()
    def acquire_parameters(self):
        """ Acquire parameters of the osciloscope.
            It is intended for adjusting the values obtained in acquire_curve
        """
        values = 'XZE?;XIN?;PT_OF?;YZE?;YMU?;YOF?;'
        answer = self.query('WFMP:{}'.format(values))
        parameters = {}
        for v, j in zip(values.split('?;'),answer.split(';')):
            parameters[v] = float(j)
        return parameters
    
    @Action()
    def data_setup(self):
        """ Sets the way data is going to be encoded for sending. 
        """
        self.send('DAT:ENC ASCI;WID 2') #ASCII is the least efficient way, but
                                        # couldn't make the binary mode to work

    @Action()
    def acquire_curve(self,start=1,stop=2500):
        """ Gets data from the oscilloscope. It accepts setting the start and 
            stop points of the acquisition (by default the entire range).
        """
        parameters = self.acquire_parameters()
        self.data_setup() 
        self.send('DAT:STAR {}'.format(start))
        self.send('DAT:STOP {}'.format(stop))
        data = self.query('CURV?')
        data = data.split(',')
        data = np.array(list(map(float,data)))
        ydata = (data - parameters['YOF']) * parameters['YMU']\
                + parameters['YZE']
        xdata = np.arange(len(data))*parameters['XIN'] + parameters['XZE']
        return list(xdata), list(ydata)
        
    
    @Action()
    def forcetrigger(self):
        """ Creates a trigger event. 
        """
        self.send('TRIG:FORC')
        return
        
    @Action()
    def triggerlevel(self):
        """ Sets the trigger level to 50% of the minimum and maximum values of 
            the signal. 
        """
        self.send('TRIG:MAI SETL')
    
    @Feat(values={'AUTO', 'NORMAL'})
    def trigger(self):
        """ Retrieves trigger state.
        """
        return self.query('TRIG:MAIN:MODE?')
    
    @trigger.setter
    def trigger(self,state):
        """ Sets the trigger state.
        """
        self.send('TRIG:MAI:MOD {}'.format(state))
        return
        
    @Feat()
    def horizontal_division(self):
        """ Horizontal time base division. 
        """
        return float(self.query('HOR:MAI:SCA?'))
    
    @horizontal_division.setter
    def horizontal_division(self,value):
        """ Sets the horizontal time base division. 
        """
        self.send('HOR:MAI:SCA {}'.format(value))
        return
        
    @Feat(values={0, 4, 16, 64, 128})
    def number_averages(self):
        """ Number of averages
        """
        answer = self.query('ACQ?')
        answer = answer.split(';')
        if answer[0] == 'SAMPLE':
            return 0
        elif answer[0] == 'AVERAGE':
            return int(self.query('ACQ:NUMAV?'))
        else:
            raise InvalidCommand
    
    @number_averages.setter
    def number_averages(self,value):
        """ Sets the number of averages. If 0, the it is a continous sample.
        """
        if value == 0:
            self.send('ACQ:MOD SAMPLE')
        else:
            self.send('ACQ:MOD AVE;NUMAV {}'.format(value))
        
    @Action(values={'FREQ', 'MINI', 'MAXI', 'MEAN'})
    def _measure(self, mode):
        """ Measures the Frequency, Minimum, Maximum or Mean of a signal.
        """
        self.send('MEASU:IMM:TYP {}'.format(mode))
        return float(self.query('MEASU:IMM:VAL?'))
    
    def measure_mean(self):
        """ Gets the mean of the signal.
        """
        answer = self._measure('MEAN')
        return answer
 
    def measure_frequency(self):
        """ Gets the frequency of the signal.
        """
        answer = self._measure('FREQ')
        return answer
        
    def measure_minimum(self):
        """ Gets the minimum of the signal.
        """
        answer = self._measure('MINI')
        return answer        
       
    def measure_maximum(self):
        """ Gets the mean of the signal.
        """
        answer = self._measure('MAXI')
        return answer
        
if __name__ == '__main__':
    import argparse
    
    parser = argparse.ArgumentParser(description='Measure using TDS1012 and dump to screen')
    parser.add_argument('-p', '--port', default='/dev/ttyS0',
                        help='Serial port')
    parser.add_argument('-v', '--view', action='store_true', default=True,
                        help='View ')
    parser.add_argument('-c', '--channel', default=1, type=int,
                        help='Channel to use')
   

    args = parser.parse_args()

    osc = TDS1012(args.port)
    osc.initiate()
    print('Osciloscope Identification: {}'.format(osc.idn))
    print(osc.trigger)
    osc.forcetrigger()
    osc.triggerlevel()
    osc.trigger = "AUTO"
    print(osc.trigger)

    params = osc.acquire_parameters()

    if args.view:
        import matplotlib.pyplot as plt
        
    if args.view:
        osc.datasource = args.channel
        x, y = osc.acquire_curve()
        x = np.array(x)
        x = x - x.min()
        y = np.array(y)
        plt.plot(x, y)
        plt.show()



# -*- coding: utf-8 -*-
"""
    lantz.drivers.kepco.bop
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation of Kepco BOP power supply
    Author: Kevin Miao
    Date: 1/18/2016
"""

from lantz import Action, Feat
from lantz.messagebased import MessageBasedDriver

class BOP(MessageBasedDriver):

    @Feat()
    def idn(self):
        return self.query('*IDN?')

    @Action(values={'curr', 'volt'})
    def mode(self, value):
        """
        sets operation mode
        curr - constant current mode
        volt - constant voltage mode
        """
        return self.write('FUNC:MODE {}'.format(value))

    @Feat(units='A')
    def current(self):
        return self.query('MEAS:CURR?')

    @current.setter
    def current(self, value):
        self.write('CURR {:1.1f}'.format(value))

    @Feat(units='V')
    def voltage(self):
        return self.query('MEAS:VOLT?')

    @voltage.setter
    def voltage(self, value):
        self.write('VOLT {:1.1f}'.format(value))



"""
    lantz.drivers.keysight.e8364b
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Drivers for the E8364B Network Analyzer using RAW SOCKETS

    Authors: Alexandre Bourassa
    Date: 24/03/2016
"""
import numpy as _np

from lantz import Feat, DictFeat, Action
from lantz.feat import MISSING
from lantz.errors import InstrumentError
from lantz.messagebased import MessageBasedDriver

class E8364B(MessageBasedDriver):
    """E8364B Network Analyzer
    """


    DEFAULTS = {'COMMON': {'write_termination': '\n',
                           'read_termination': '\n',
                           'timeout':10000,}}

    @Feat(read_once=True)
    def idn(self):
        return self.query('*IDN?')

# ----------------------------------------------------
#       Sweep Settings functions
# ----------------------------------------------------

    @Feat(units='Hz')
    def center_frequency(self):
        """Center Frequency
        """
        return float(self.query('SENS:FREQ:CENT?'))

    @center_frequency.setter
    def center_frequency(self, value):
        self.write('SENS:FREQ:CENT {}'.format(int(value)))


    @Feat(units='Hz')
    def span(self):
        """Span of the sweep
        """
        return float(self.query('SENS:FREQ:SPAN?'))

    @span.setter
    def span(self, value):
        self.write('SENS:FREQ:SPAN {}'.format(int(value)))

    @Feat(units='Hz')
    def if_bandwidth(self):
        """Bandwidth of the digital IF filter
        """
        return float(self.query('SENS:BAND?'))

    @if_bandwidth.setter
    def if_bandwidth(self, value):
        self.write('SENS:BAND {}'.format(int(value)))

    @Feat(limits=(1, 16001, 1))
    def nb_points(self):
        """The number of data points for the measurement
        """
        return int(self.query('SENS:SWE:POIN?'))

    @nb_points.setter
    def nb_points(self, points):
        self.write('SENS:SWE:POIN {}'.format(int(points)))

    ALLOWED_MEAS_TYPE = {"S11", "S12", "S21", "S22"}
    @Feat(values=ALLOWED_MEAS_TYPE)
    def meas_type(self):
        return self.get_measurement_catalog()['CH1_S11_1']

    @meas_type.setter
    def meas_type(self, meas_type):
        self.write("CALC:PAR:MOD {}".format(meas_type))


# ----------------------------------------------------
#       Power functions
# ----------------------------------------------------

    @Feat(values={True: 1, False: 0})
    def power_on(self):
        """RF Power State (True==ON or False==OFF)
        """
        return int(self.query('OUTP?'))

    @power_on.setter
    def power_on(self, state=True):
        self.write('OUTP {}'.format(state))

    @Feat()
    def power_level(self):
        """RF Power level in dBm
        """
        return float(self.query("SOUR:POW?"))

    @power_level.setter
    def power_level(self, level):
        self.write("SOUR:POW {}".format(level))

    ## Possibly not available on this model...
    # @Feat(values={True: 1, False: 0})
    # def manual_noise_on(self):
    #     """RF Power State (True==ON or False==OFF)
    #     """
    #     return int(self.query('OUTP:MAN:NOIS?'))
    #
    # @manual_noise_on.setter
    # def manual_noise_on(self, state=True):
    #     if state:   state_str = 'ON'
    #     else:       state_str = 'OFF'
    #     self.write('OUTP:MAN:NOIS ' + state_str)

#----------------------------------------------------
#       Averaging functions
#----------------------------------------------------

    @Feat(values={True: 1, False: 0})
    def average_on(self):
        """Averaging state (True==ON or False==OFF)
        """
        return int(self.query('SENS:AVER:STAT?'))

    @average_on.setter
    def average_on(self, state=True):
        self.write('SENS:AVER:STAT {}'.format(state))

    @Feat(limits=(1,65536, 1))
    def average_count(self):
        """Averaging count
        """
        return int(self.query('SENS:AVER:COUN?'))

    @average_count.setter
    def average_count(self, counts):
        self.write('SENS:AVER:COUN {}'.format(int(counts)))

    @Action()
    def clear_average(self):
        """Reset the averaging
        """
        self.write('SENS:AVER:CLE')



#----------------------------------------------------
#       Data Query functions
#----------------------------------------------------
    @Feat(values={"REAL32":"REAL,+32", "REAL64":"REAL,+64", "ASCII":"ASC,+0"})
    def data_format(self):
        return self.query("FORM:DATA?")

    @data_format.setter
    def data_format(self, value):
        self.write('FORM:DATA {}'.format(value))

    def query_data(self, command, use_cached=True):
        # For quick data acquisition, let's assume data format as not changed since last query
        form = self.recall('data_format') if use_cached else self.data_format
        if form is MISSING: return _np.array([])
        if form == "REAL64":
            return _np.array(self.resource.query_binary_values(command, datatype='d', is_big_endian=True))
        elif form == "REAL32":
            return _np.array(self.resource.query_binary_values(command, datatype='f', is_big_endian=True))
        elif form == "ASCII":
            return _np.array(list(map(float, self.query(command).split(','))))
        else:
            raise Exception(str(form) + "Invalid data format")

    @Action()
    def x_data(self):
        return self.query_data('SENS:X?')

    @Action()
    def y_data(self):
        data = self.query_data("CALC:DATA? SDATA")
        data = data[::2]+data[1::2]*1j
        return data


# ----------------------------------------------------
#       Traces and Channel functions (Complicates the remote use of the device
#       so don't use unless you know what you are doing...)
#       For simplicity, we will always use CH1 and parameter name 'CH1_S11_1'
# ----------------------------------------------------
    def clear_all_traces(self):
        self.write('SYST:FPR')
        self.write('DISP:WIND:STAT ON')

    def create_new_measurement(self, name='CH1_S11_1', meas_type="S11"):
        if not meas_type in self.ALLOWED_MEAS_TYPE: raise Exception("Invalid meas_type: "+str(meas_type))
        self.write("CALC:PAR:DEF '"+name+"',"+meas_type)
        self.write("DISP:WIND:TRAC:FEED '"+name+"'")


    def get_measurement_catalog(self):
        response = self.query("CALC:PAR:CAT?").translate({ord('"'):None}).split(',')
        ans = dict()
        for i in range(int(len(response)/2)):
            ans[response[2*i]]=response[2*i+1]
        return ans

    def select_measurement(self, name):
        self.write("CALC:PAR:SEL '{}'".format(name))

# ----------------------------------------------------
#       Misc. functions
# ----------------------------------------------------

    @Action()
    def initialize(self):
        super(E8364B, self).initialize()
        meas = self.get_measurement_catalog()
        if not 'CH1_S11_1' in meas:
            self.clear_all_traces()
            self.create_new_measurement(name='CH1_S11_1', meas_type="S11")

        self.select_measurement('CH1_S11_1')
        self.data_format = 'REAL64'


















#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

import os
import codecs


def read(filename):
    return codecs.open(filename, encoding='utf-8').read()


long_description = '\n\n'.join([read('README'),
                                read('AUTHORS'),
                                read('CHANGES')])

__doc__ = long_description

root_folder = os.path.dirname(os.path.abspath(__file__))

# Compile a list of companies with drivers.
folder = os.path.join(root_folder, 'lantz', 'drivers')
paths = os.listdir(folder)
companies = [path for path in paths
             if os.path.isdir(os.path.join(folder, path))
             and os.path.exists(os.path.join(folder, path, '__init__.py'))]


setup(name='lantz-drivers',
      version='0.5.2',
      license='BSD',
      description='Instrumentation framework',
      long_description=long_description,
      keywords='measurement control instrumentation science',
      author='Hernan E. Grecco',
      author_email='hernan.grecco@gmail.com',
      url='https://github.com/lantzproject',
      packages=['lantz.drivers'] + ['lantz.drivers.' + company for company in companies],
      install_requires=['lantz-core>=0.5',
                        ],
      zip_safe=False,
      platforms='any',
      classifiers=[
           'Development Status :: 4 - Beta',
           'Intended Audience :: Developers',
           'Intended Audience :: Science/Research',
           'License :: OSI Approved :: BSD License',
           'Operating System :: MacOS :: MacOS X',
           'Operating System :: Microsoft :: Windows',
           'Operating System :: POSIX',
           'Programming Language :: Python',
           'Programming Language :: Python :: 3.2',
           'Programming Language :: Python :: 3.3',
           'Programming Language :: Python :: 3.4',
           'Topic :: Scientific/Engineering',
           'Topic :: Software Development :: Libraries'
      ],
)



# -*- coding: utf-8 -*-
"""
    lantz.drivers.pi
    ~~~~~~~~~~~~~~~~~~~~~~

    :company: PhysikInstrumente
    :description: Motion and positioning components
    :website: https://www.physikinstrumente.com

    ----

    :copyright: 2017 by Vasco Tenner
    :license: BSD, see LICENSE for more details.
"""

from .piezo import Piezo, parse_line, parse_multi

__all__ = ['Piezo','parse_line','parse_multi']



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.newport.powermeter1830c
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the drivers to control an Optical Power Meter.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
    
    Source: Instruction Manual (Newport)
"""

from lantz.feat import Feat
from lantz.action import Action
from lantz.drivers.legacy.serial import SerialDriver


class powermeter1830c(SerialDriver):
    """ Newport 1830c Power Meter
    """
    
    ENCODING = 'ascii'
    RECV_TERMINATION = '\n'
    SEND_TERMINATION = '\n'
    TIMEOUT = 2000   

    
    def __init__(self, port=1,baudrate=9600):
        super().__init__(port, baudrate, bytesize=8, parity='None', stopbits=1) 
    
    def initialize(self):
        super().initialize() 
        
    @Feat(values={True: 1, False: 0})
    def attenuator(self):
        """ Attenuator. 
            1: Attenuator present
            0: Attenuator not present
        """
        return int(self.query('A?'))
    
    @attenuator.setter
    def attenuator(self, value):
        self.send('A{}'.format(value))
        
    @Feat(values={True: 1, False: 0})
    def beeper(self):
        """ Checks whether the audio output is on or off.
        """
        return int(self.query('B?'))
    
    @beeper.setter
    def beeper(self,value):
        self.send('B{}'.format(value))
        
    @Feat
    def data(self):
        """ Retrieves the value from the power meter.
        """        
        return float(self.query('D?'))
    
    @Feat(values={True: 1, False: 0})
    def echo(self):
        """ Returns echo mode. Only applied to RS232 communication
        """
        return int(self.query('E?'))
    
    @echo.setter
    def echo(self,value):
        self.send('E{}'.format(value))
        
    @Feat(values={'Slow': 1, 'Medium': 2, 'Fast': 3})
    def filter(self):
        """ How many measurements are averaged for the displayed reading.
            slow: 16 measurements
            medium: 4 measurements
            fast: 1 measurement. 
        """
        return int(self.query('F?'))
    
    @filter.setter
    def filter(self,value):
        self.send('F{}'.format(value))
        
    @Feat(values={True: 1, False: 0})
    def go(self):
        """ Enable or disable the power meter from taking new measurements. 
        """
        return int(self.query('G?'))
    
    @go.setter
    def go(self,value):
        self.send('G{}'.format(value))
        
    @Feat(values={'Off': 0, 'Medium': 1, 'High': 2})
    def keypad(self):
        """ Keypad/Display backlight intensity levels.
        """
        return int(self.query('K?'))
    
    @keypad.setter
    def keypad(self,value):
        self.send('K{}'.format(value))
        
    @Feat(values={True: 1, False: 0})
    def lockout(self):
        """ Enable/Disable the lockout. When the lockout is enabled, any front panel key presses would have no effect on system operation.
        """
        return int(self.query('L?'))
    
    @lockout.setter
    def lockout(self,value):
        self.send('L{}'.format(value))

    @Action()
    def autocalibration(self):
        """ Autocalibration of the power meter. This procedure disconnects the input signal. 
            It should be performed at least 60 minutes after warm-up.
        """
        self.send('O')
    
    @Feat(values=set(range(0,9)))
    def range(self):
        """ Set the signal range for the input signal. 
            0 means auto setting the range. 1 is the lowest signal range and 8 the highest.
        """
        return int(self.query('R?'))
    
    @range.setter
    def range(self,value):
        self.send('R{}'.format(value))
        
    @Action()
    def store_reference(self):
        """ Sets the current input signal power level as the power reference level. 
            Each time the S command is sent, the current input signal becomes the new reference level. 
        """
        self.send('S')
        
    @Feat(values={'Watts': 1, 'dB': 2, 'dBm': 3, 'REL': 4})
    def units(self):
        """ Sets and gets the units of the measurements. 
        """ 
        return int(self.query('U?'))
    
    @units.setter
    def units(self,value):
        self.send('U{}'.format(value))
        
    @Feat(limits=(1,10000,1))
    def wavelength(self):
        """ Sets and gets the wavelength of the input signal.
        """
        return int(self.query('W?'))
    
    @wavelength.setter
    def wavelength(self,value):
        self.send('W{}'.format(int(value)))
    
    @Feat(values={True: 1, False: 0})
    def zero(self):
        """ Turn the zero function on/off. Zero function is used for subtracting any background power levels in future measurements. 
        """
        return int(self.query('Z?'))
    
    @zero.setter
    def zero(self,value):
        self.send('Z{}'.format(value))
        
if __name__ == '__main__':
    import argparse
    import lantz.log
    
    parser = argparse.ArgumentParser(description='Test Kentech HRI')
    parser.add_argument('-p', '--port', type=str, default='1',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantz.log.log_to_socket(lantz.log.DEBUG)
    
    with powermeter1830c(args.port) as inst:
        
        inst.initialize() # Initialize the communication with the power meter
        
        inst.lockout = True # Blocks the front panel
        inst.keypad = 'Off' # Switches the keypad off
        inst.attenuator = True # The attenuator is on
        inst.wavelength = 633 # Sets the wavelength to 633nm
        inst.units = "Watts" # Sets the units to Watts
        inst.filter = 'Slow' # Averages 16 measurements
        
        if not inst.go:
            inst.go = True # If the instrument is not running, enables it
            
        inst.range = 0 # Auto-sets the range
        
        print('The measured power is {} Watts'.format(inst.data))



# -*- coding: utf-8 -*-
"""
    lantz.drivers.example.foreign_example
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Foreign library example.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import ctypes as ct

from lantz import Feat, Action, DictFeat
from lantz.foreign import LibraryDriver
from lantz import errors

class ForeignTemplate(LibraryDriver):
    """Template for Drivers using a library.
    """

    LIBRARY_NAME = 'mylibrary.dll'

    def _return_handler(self, func_name, ret_value):
        if ret_value != 0:
            raise errors.InstrumentError('{} ({})'.format(ret_value, _ERRORS[ret_value]))
        return ret_value

    @Feat()
    def idn(self):
        return self.query('*IDN?')

    @Feat(units='V', limits=(10,))
    def amplitude(self):
        """Amplitude.
        """
        return float(self.query('?AMP'))

    @amplitude.setter
    def amplitude(self, value):
        self.query('!AMP {:.1f}'.format(value))

    @DictFeat(values={True: '1', False: '0'}, keys=list(range(1,9)))
    def dout(self, key):
        """Digital output state.
        """
        return self.query('?DOU {}'.format(key))

    @dout.setter
    def dout(self, key, value):
        self.query('!DOU {} {}'.format(key, value))

    @Action()
    def do_something(self):
        """Help for do_something
        """
        return self.lib.something()



if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test Kentech HRI')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False, help='Show interactive GUI')

    args = parser.parse_args()
    lantz.log.log_to_socket(lantz.log.DEBUG)
    with ForeignTemplate() as inst:
        if args.interactive:
            from lantz.ui.app import start_test_app
            start_test_app(inst)
        else:
            # Add your test code here
            print('Non interactive mode')




# -*- coding: utf-8 -*-
"""
    lantz.drivers.ni
    ~~~~~~~~~~~~~~~~

    :company: National Instruments
    :description: 
    :website: http://www.ni.com/

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from . import daqmx




# -*- coding: utf-8 -*-
"""
    lantz.drivers.examples.voltmeter
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the Simple Voltmeter described in the Lantz tutorial.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz import Feat, DictFeat, Action
from lantz import errors
from lantz import MessageBasedDriver


class LantzVoltmeter(MessageBasedDriver):
    """Lantz Signal Generator
    """

    DEFAULTS = {'COMMON': {'write_termination': '\n',
                           'read_termination': '\n'}}

    def query(self, command, *, send_args=(None, None), recv_args=(None, None)):
        answer = super().query(command, send_args=send_args, recv_args=recv_args)
        if answer == 'ERROR':
            raise errors.InstrumentError
        return answer

    @Feat()
    def idn(self):
        return self.query('?IDN')

    @DictFeat(units='V', keys=(0, 1))
    def voltage(self, key):
        """Measure the voltage.
        """
        return float(self.query('?MEAS {}'.format(key)))

    @DictFeat(values={0.1: '0', 1: '1', 10: '2', 100: '3', 1000: '4'}, keys=(0, 1))
    def range(self, key):
        return self.query('?RANGE {}'.format(key))

    @range
    def range(self, key, value):
        self.query('!RANGE {} {}'.format(key, value))

    @Action()
    def auto_range(self, key):
        """Autoselect a range.
        """
        self.query('!ARANGE {}'.format(key))

    @Action()
    def calibrate(self):
        """Calibrate.
        """
        self.query('!CAL')

    @Action()
    def self_test(self):
        """Self test
        """
        self.query('!TES')



from lantz.messagebased import MessageBasedDriver
from lantz import Feat, DictFeat, Action
from collections import OrderedDict


#from lantz import Q_
import numpy as np


import socket
import warnings

class Ag81130A(MessageBasedDriver):
    """
    Lantz driver for interfacing with Agilent 81130A pulse pattern generator.

    Includes testing code, which should work out of the box assuming you give
    it the correct GPIB address.

    Author: P. Mintun
    Date: 12/6/2016
    Version: 0.1
    """

    DEFAULTS = {
        'COMMON': {
            'write_termination': '\n',
            'read_termination': '\n',
        }
    }

    ON_OFF_VALS = OrderedDict([
                    ('on', 1),
                    ('off', 0),
    ])

    ARM_SOURCE_VALS = OrderedDict([
                      ('immediate', 'IMM'),
                      ('external', 'EXT'),
                      ('manual', 'MAN')
    ])

    TRIG_SOURCE_VALS = OrderedDict([
                      ('immediate', 'IMM'),
                      ('external', 'EXT'),
                      ('internal', '1')
    ])

    TRIG_MODE_VALS = OrderedDict([
                      ('continuous', 'CONT'),
                      ('start', 'STAR')
    ])

    channels = range(1,3)
    #channels = OrderedDict([
    #               ('1', 0),
    #               ('2', 1)
    #               ])

    segments = range(1,5)

    # some weird list comprehension variable scope thing here
    chan_segs = [(x,y) for x in range(1,3) for y in range(1,5)]


    @Feat()
    def idn(self):
        """
        Identifiies the instrument.
        """
        return self.query('*IDN?')

    @Action()
    def reset(self):
        """
        Resets the instrument to default settings. This is recommended by the
        manual before starting to programming it.
        """
        return self.write('*RST')

    @DictFeat(keys=channels, limits=(-4.0,4.0))
    def volt_high(self, chan):
        """
        Returns the voltage corresponding to HIGH for channel chan.

        """
        return self.query('VOLT{}:HIGH?'.format(chan))

    @volt_high.setter
    def volt_high(self, chan, volts):
        """
        Sets the voltage corresponding to HIGH for channel chan to volts.
        """
        return self.write('VOLT{}:HIGH {}V'.format(chan, volts))

    @DictFeat(keys=channels, limits=(-4.0,4.0))
    def volt_low(self, chan):
        """
        Returns the voltage corresponding to LOW for channel chan.

        """
        return self.query('VOLT{}:LOW?'.format(chan))

    @volt_low.setter
    def volt_low(self, chan, volts):
        """
        Sets the voltage corresponding to LOW for channel chan to volts.
        """
        return self.write('VOLT{}:LOW {}V'.format(chan, volts))

    @Feat(values=ON_OFF_VALS)
    def display(self):
        """
        Returns if display is on or off, (off enables faster programming).
        """
        return int(self.query('DISP?'))

    @display.setter
    def display(self, on_off):
        """
        Sets display to be on or off, (off enables faster programming).
        """
        return self.write('DISP {}'.format(on_off))


    @Feat(values=ON_OFF_VALS)
    def pattern_mode(self):
        """
        Returns whether or not pattern mode is enabled.
        """
        return int(self.query('DIG:PATT?'))

    @pattern_mode.setter
    def pattern_mode(self, on_off):
        """
        Sets pattern mode to be enabled or disabled.
        """
        return self.write('DIG:PATT {}'.format(on_off))

    @Feat(values=ARM_SOURCE_VALS)
    def arm_source(self):
        """
        Returns the source used for the arming signal for triggering the instrument.

        Options are immediate (continuous mode), external trigger, and manually
        triggered from the keypad.
        """
        return self.query('ARM:SOUR?')

    @arm_source.setter
    def arm_source(self, source_channel):
        """
        Sets the trigger signal to the source channel.

        Options are immediate (continuous mode), external trigger, and manually
        triggered from the keypad.
        """
        return self.write('ARM:SOUR {}'.format(source_channel))

    @Feat(values=TRIG_SOURCE_VALS)
    def trigger_source(self):
        """
        Returns the source of the pulse period trigger signal.

        Options are immediate, internal, or external (CLK IN signal)
        """
        return self.query('TRIG:SOUR?')

    @trigger_source.setter
    def trigger_source(self, trigger_source):
        """
        Sets the source of the pulse period trigger signal.

        Options are immediate, internal, or external (CLK IN signal)
        """
        return self.write('TRIG:SOUR {}'.format(trigger_source))

    @DictFeat(keys=segments)
    def dig_patt_length(self, seg_num, limits=(0,65504,1)):
        """
        Returns the segment type
        """

        return int(self.query('DIG:PATT:SEGM{}:LENG?'.format(seg_num)))

    @dig_patt_length.setter
    def dig_patt_length(self, seg_num, length):

        return self.write('DIG:PATT:SEGM{}:LENG {}'.format(seg_num, int(length)))

    @DictFeat(keys=chan_segs, values={'data':'DATA', 'PRBS':'PRBS', 'high':'HIGH', 'low':'LOW'})
    def dig_patt_type(self, chan_seg):
        """
        Returns the segment type
        """
        channel = chan_seg[0]
        seg_num = chan_seg[1]

        return self.query('DIG:PATT:SEGM{}:TYPE{}?'.format(seg_num, channel))

    @dig_patt_type.setter
    def dig_patt_type(self, chan_seg, patt_type):

        channel = chan_seg[0]
        seg_num = chan_seg[1]

        return self.write('DIG:PATT:SEGM{}:TYPE{} {}'.format(seg_num, channel, patt_type))

    @Feat(limits=(1e3,660e6))
    def frequency(self):
        """
        Gets the operating frequency of the device - this is what sets the timescale
        of the pattern duration.
        """
        return float(self.query('FREQ?'))

    @frequency.setter
    def frequency(self, Hz):
        """
        Sets the internal PLL frequency to Hz.
        """
        return self.write('FREQ {}'.format(Hz))

    @DictFeat(keys=channels, values={'nrz':'NRZ', 'rz':'RZ', 'r1':'R1'})
    def data_format(self, channel):
        """
        Returns current data format for the given channel.

        Options are:
        - nrz (non-return to zero)
        - rz (return to zero)
        - r1 (?)
        """
        return self.query('DIG:SIGN{}:FORM?'.format(channel))

    @data_format.setter
    def data_format(self, channel, data_format):
        """
        Sets data format of the given channel to data_format.

        Options are:
        - nrz (non-return to zero)
        - rz (return to zero)
        - r1 (?)
        """
        return self.write('DIG:SIGN{}:FORM {}'.format(channel, data_format))

    @DictFeat(keys=channels, values=ON_OFF_VALS)
    def output_on(self, channel):
        """
        Queries the output of the specified channel.
        """
        return int(self.query('OUTP{}?'.format(channel)))

    @output_on.setter
    def output_on(self, channel, state):
        """
        Sets the output of the specified channel to state.
        """
        return self.write('OUTP{} {}'.format(channel, state))

    @DictFeat(keys=channels, values=ON_OFF_VALS)
    def comp_output_on(self, channel):
        """
        Queries the output of the specified channel.
        """
        return int(self.query('OUTP{}:COMP ?'.format(channel)))

    @comp_output_on.setter
    def comp_output_on(self, channel, state):
        """
        Sets the output of the specified channel to state.
        """
        return self.write('OUTP{}:COMP {}'.format(channel, state))

    @DictFeat(keys=chan_segs)
    def segment_data(self, chan_seg):
        """
        Returns the data from segment seg_num, channel chan
        """
        channel = chan_seg[0]
        seg_num = chan_seg[1]
        result = self.query('DIG:PATT:SEGM{}:DATA{}?'.format(seg_num, channel))

        # now process data
        return result

    @segment_data.setter
    def segment_data(self, chan_seg, data_stream):
        """
        Sets the data from segment seg_num, channel chan to data_stream (numpy array)
        """
        print('called data setter')


        channel = chan_seg[0]
        seg_num = chan_seg[1]

        data = self.encode_data(data_stream[0])

        self.write('DIG:PATT:SEGM{}:DATA{} {}'.format(seg_num, channel, data))

        return data



    @Feat(limits=(1,5,1))
    def start_seg(self):
        """
        Queries the starting segment for the device pattern.
        """
        return int(self.query('DIG:PATT:LOOP:STAR?'))

    @start_seg.setter
    def start_seg(self, segment):
        """
        Sets the starting segment for the device pattern to segment.
        """
        return self.write('DIG:PATT:LOOP:STAR {}'.format(segment))

    @Feat(limits=(1,5,1))
    def loop_length(self):
        """
        Queries the number of segments to be repeated in the loop.
        """
        return int(self.query('DIG:PATT:LOOP:LENG?'))

    @loop_length.setter
    def loop_length(self, length):
        """
        Sets the number of segments to be included in the loop.
        """
        return self.write('DIG:PATT:LOOP:LENG {}'.format(length))

    @DictFeat(keys=channels, limits=(0, 2*np.pi))
    def phase_delay(self, chan):
        """
        Returns the phase delay of the output signal of channel chan, in radians.
        """
        return float(self.query('PHAS{}?'.format(chan)))

    @phase_delay.setter
    def phase_delay(self, chan, delay):
        """
        Sets the phase delay of the output signal of channel chan to delay, in radians.
        """
        return self.write('PHAS{} {}'.format(chan, delay))

    @DictFeat(keys=channels, limits=(0, 3000e-9))
    def timed_delay(self, chan):
        """
        Returns the timed delay of the output signal of channel chan in seconds.
        """
        return float(self.query('PULS:DEL{}?'.format(chan)))


    @timed_delay.setter
    def timed_delay(self, chan, sec):
        """
        Sets the timed delay of output of channel chan to sec.
        """
        return self.write('PULS:DEL{} {}S'.format(chan, sec))


    @Feat()
    def trig_output(self):
        """
        Returns the voltage level used for trigger output.
        """
        return self.query('PULS:TRIG:VOLT?')

    @Feat()
    def trig_pos(self):
        """
        Returns the trigger out position in pattern mode, returning the segment number.
        """
        return self.query('PULS:TRIG:POS?')

    @Feat(values=TRIG_MODE_VALS)
    def trig_mode(self):
        """
        Returns the trigger out generation mode in pattern mode.
        """
        return self.query('PULS:TRIG:MODE?')

    @trig_mode.setter
    def trig_mode(self, trigger_mode):
        """
        Sets the trigger out generation mode (pattern mode only). Options are
        continuous or start.
        """
        return self.write('PULS:TRIG:MODE {}'.format(trigger_mode))


    def encode_data(self, data_series):
        """
        Helper function to implement IEEE 488.2 7.7.6.2 program data protocol.

        Encodes data_series (numpy byte array) into format that can be read by PPG.
        """
        # starts with # character
        data_string = '#'

        # hack to avoid issues w/ ellipses in large np arrays
        np.set_printoptions(threshold=65536)

        raw_data = np.array_str(data_series, max_line_width=65536)

        np.set_printoptions(threshold=1000)
        # figure out length of data_series
        data_length = data_series.size
        # figure out length of length of data_series
        len_data_length = len(str(data_length))

        # add all this stuff
        data_string += str(len_data_length)
        data_string += str(data_length)

        # TODO: fix import
        #max_line_width avoids adding newline or whitespace
        #raw_data = np.array_str(data_series, max_line_width=1000000)

        data_string += raw_data[1:-1:2] #strips out left bracket, right bracket, and spaces

        return data_string

    def decode_data(self, encoded_series):
        """
        Helper function to implement IEEE 488.2 7.7.6.2 program data protocol.

        Decodes encoded_series from PPG into raw data that can be read.
        """

        if encoded_series[0] != '#':

            print('invalid encoded series!')

        len_len = int(encoded_series[1])

        char_list = list(encoded_series[2+len_len:])

        return [int(x) for x in char_list]


    def preview_wfm(self):

        import matplotlib.pyplot as plt

        # code to figure out timeseries + number of points
        loop_start = self.start_seg
        loop_len = self.loop_length

        segments = [loop_start]
        current_seg = loop_start

        # probably not the best way possible to do this, but it works...
        while loop_len > 1:
            current_seg += 1
            loop_len -= 1

            if current_seg > 4:
                segments.append(current_seg % 4)

            else:
                segments.append(current_seg)

        patt_length = 0

        for seg in segments:
            patt_length += self.dig_patt_length[seg]
            print('seg{}:{}'.format(seg, self.dig_patt_length[seg]))

        print('Total length:{}'.format(patt_length))

        freq  = self.frequency
        t = np.arange(0, patt_length, 1)/freq

        chan1 = np.zeros(patt_length)
        chan2 = np.zeros(patt_length)

        current_index = 0

        for seg in segments:

            seg_type_1 = self.dig_patt_type[(1, seg)]
            seg_type_2 = self.dig_patt_type[(2, seg)]
            length = self.dig_patt_length[seg]

            if seg_type_1 == 'low':
                chan1[current_index:current_index+length] = 0

            elif seg_type_1 == 'high':
                chan1[current_index:current_index+length] = 1

            elif seg_type_1 == 'data':
                chan1[current_index:current_index+length] = self.segment_data[(1,seg)]


            if seg_type_2 == 'low':
                chan2[current_index:current_index+length] = 0

            elif seg_type_2 == 'high':
                chan2[current_index:current_index+length] = 1

            elif seg_type_2 == 'data':
                chan2[current_index:current_index+length] = self.segment_data[(2, seg)]

            current_index += length
            #chan1 = np.zeros()
            #chan2 = np

        def square(t_val, tmax):
            """
            Square wave helper function for plotting trigger output
            """

            if t_val < tmax/2.0:
                return 1
            else:
                return 0
        vectorized_square = np.vectorize(square) #vectorize because :gottagofast:

        plt.figure(1)
        plt.subplot(311)
        plt.ylabel('$T_0$')
        axes = plt.gca()
        axes.step(t, vectorized_square(t, t.max()), 'k-', where='mid')
        axes.set_ylim([-0.5,1.5])

        # now plot series from channel 1
        plt.subplot(312)
        plt.ylabel('Channel 1')
        axes = plt.gca()
        axes.step(t, chan1, 'r--', where='mid')
        axes.set_ylim([-0.5,1.5])

        # plot series from channel 2
        plt.subplot(313)
        plt.ylabel('Channel 2')
        axes = plt.gca()
        axes.step(t, chan2, 'r--', where='mid')
        axes.set_ylim([-0.5,1.5])
        #plt.show()

    @Action()
    def odmr_waveform(self, preview_wfm=False, ref_freq=503.0):

        ref_freq = 503.0
        print('Setting up ODMR waveforms')

        print('Identification: {}'.format(self.idn))
        self.reset() #initializes default parameters for clean setup
        self.display = 'off'
        print('Display off?: {}'.format(self.display))

        self.pattern_mode = 'on'
        print('Digital pattern mode on?:{}'.format(self.pattern_mode))

        self.arm_source = 'immediate' # sets continuous operation
        print('Arm source immediate?: {}'.format(self.arm_source))

        # output TTL pulses for RF switch on channel 1
        # TTL pulses should be between 0 (low) and 2.5 (high) volts
        # so set up channel 1 output like this
        self.volt_low[1] = 0.0
        self.volt_high[1] = 2.5

        print('High voltage, should be 2.5 V:{}'.format(self.volt_high[1]))
        print('Low voltage, should be 0 V:{}'.format(self.volt_low[1]))


        self.volt_low[2] = 0.0
        self.volt_high[2] = 1.0

        print('High voltage, should be 1.0 V:{}'.format(self.volt_high[2]))
        print('Low voltage, should be 0 V:{}'.format(self.volt_low[2]))

        self.data_format[1] = 'nrz'
        self.data_format[2] = 'nrz'


        #ref_freq = 503.0 #Hz
        # since we have two pieces to the square wave, the frequency generator should
        # be set to use at least twice the references.

        self.frequency = 2*ref_freq
        self.dig_patt_length[1] = 4

        self.dig_patt_type[(2,1)] = 'high'
        self.dig_patt_type[(1,1)] = 'data'

        self.write('DIG:PATT:SEGM1:PRES1 2, 2')

        #self.segment_data[2,1] = [np.ones(4)]

        print('Internal PLL frequency:{}'.format(self.frequency))

        #print(self.segment_data[1,1])

        #print(self.segment_data[2,1])

        #self.output_on[1] = 'on'
        #self.output_on[2] = 'on'

        print(self.output_on[1])
        print(self.output_on[2])

        self.output_on[1] = 'on'
        self.output_on[2] = 'on'

        print(self.output_on[1])
        print(self.output_on[2])

        print(self.output_on[2])
        print(self.output_on[2])


        #self.dig_patt_length[2] = scale_factor
        # ignore last two segments
        #self.dig_patt_length[3] = 0
        #self.dig_patt_length[4] = 0

        #for i in range(1,5):
        #    print('Segment {} length: {}'.format(i, self.dig_patt_length[i]))
        #if preview_wfm:
        #    self.preview_wfm()
        # configure two segements, one where channel 1 is high + one where channel 1 is low
        #self.dig_patt_type[(1, 1)] = 'high'
        #print('Channel {}, segment {} type:{}'.format(1, 1, self.dig_patt_type[(1, 1)]))
        #self.dig_patt_type[(1, 2)] = 'low'
        #print('Channel {}, segment {} type:{}'.format(1, 2, self.dig_patt_type[(1, 2)]))


        # external sync goes to TTL ref in on back of lockin
        print('TODO: check that the output on the scope of this is actually reasonable')


    def outputs_high(self, preview_wfm=False, ref_freq=503.0):

        ref_freq = 10e6
        print('PPG all outputs high!')

        print('Identification: {}'.format(self.idn))
        self.reset() #initializes default parameters for clean setup
        self.display = 'off'
        print('Display off?: {}'.format(self.display))

        self.pattern_mode = 'on'
        print('Digital pattern mode on?:{}'.format(self.pattern_mode))

        self.arm_source = 'immediate' # sets continuous operation
        print('Arm source immediate?: {}'.format(self.arm_source))

        # output TTL pulses for RF switch on channel 1
        # TTL pulses should be between 0 (low) and 2.5 (high) volts
        # so set up channel 1 output like this
        self.volt_low[1] = 0.0
        self.volt_high[1] = 2.5

        print('High voltage, should be 2.5 V:{}'.format(self.volt_high[1]))
        print('Low voltage, should be 0 V:{}'.format(self.volt_low[1]))


        self.volt_low[2] = 0.0
        self.volt_high[2] = 1.0

        print('High voltage, should be 1.0 V:{}'.format(self.volt_high[2]))
        print('Low voltage, should be 0 V:{}'.format(self.volt_low[2]))

        self.data_format[1] = 'nrz'
        self.data_format[2] = 'nrz'


        #ref_freq = 503.0 #Hz
        # since we have two pieces to the square wave, the frequency generator should
        # be set to use at least twice the references.

        self.frequency = 2*ref_freq
        self.dig_patt_length[1] = 4

        self.dig_patt_type[(2,1)] = 'high'
        self.dig_patt_type[(1,1)] = 'high'


        #self.segment_data[2,1] = [np.ones(4)]

        print('Internal PLL frequency:{}'.format(self.frequency))

        #print(self.segment_data[1,1])

        #print(self.segment_data[2,1])

        #self.output_on[1] = 'on'
        #self.output_on[2] = 'on'

        print(self.output_on[1])
        print(self.output_on[2])

        self.output_on[1] = 'on'
        self.output_on[2] = 'on'

        print(self.output_on[1])
        print(self.output_on[2])

        print(self.output_on[2])
        print(self.output_on[2])




    def rabi_waveform_step(inst, step_number):
        # helper function to program the second segment of PPG waveforms to perform Rabi
        #print('not implemented yet!')

        #inst.output_on[1] = 'off'
        #inst.output_on[2] = 'off'

        #inst.comp_output_on[1] = 'off'

        print(step_number)
        T_init = 13200
        T_rabi_max = 224
        T_readout = 13200

        off_len = T_rabi_max/2 - step_number
        on_len = 2*step_number

        data = [np.hstack((np.zeros(off_len, dtype='int'), np.ones(on_len, dtype='int'), np.zeros(off_len, dtype='int')))]
        print(data)

        #readout = [np.hstack((np.ones(T_readout + T_init, dtype='int'), np.zeros(T_rabi_max, dtype='int'), np.ones(T_readout, dtype='int')))]

        #inst.segment_data[(1,2)] = data


        encoded = inst.encode_data(data[0])

        seg_num = 2
        channel = 1

        inst.write('DIG:PATT:SEGM{}:DATA{} {}'.format(seg_num, channel, encoded))

        print('Channel {}, segment {} data:'.format(1, 2, inst.segment_data[(1,2)]))

        #inst.output_on[2] = 'on'


        return -1

    def rabi_waveform_setup(inst, rabi_params):

        # unpack measurement paramters
        #T_init = rabi_params['T_init']
        #T_readout = rabi_params['T_readout']

        print('Running Rabi waveform')

        print('Identification: {}'.format(inst.idn))
        inst.reset() #initializes default parameters for clean setup
        inst.display = 'off'
        print('Display off?: {}'.format(inst.display))

        inst.pattern_mode = 'on'
        print('Digital pattern mode on?:{}'.format(inst.pattern_mode))

        #inst.arm_source = 'immediate' # sets continuous operation
        #print('Arm source immediate?: {}'.format(inst.arm_source))



        inst.frequency = 660e6/20.0
        # output TTL pulses for RF switch on channel 1
        # TTL pulses should be between 0 (low) and 2.5 (high) volts
        # so set up channel 1 output like this
        inst.volt_low[1] = 0.0
        inst.volt_high[1] = 2.5
        print('MW TTL high voltage, should be 2.5 V:{}'.format(inst.volt_high[1]))
        print('MW TTL low voltage, should be 0 V:{}'.format(inst.volt_low[1]))

        inst.data_format[1] = 'nrz'
        inst.output_on[1] = 'on'
        inst.comp_output_on[1] = 'on' #for scope viewing


        # set up laser channel
        inst.volt_low[2] = 0.0
        inst.volt_high[2] = 1.0
        print('AOM high voltage, should be 1.0 V:{}'.format(inst.volt_high[2]))
        print('AOM Low voltage, should be 0 V:{}'.format(inst.volt_low[2]))

        inst.data_format[2] = 'nrz'
        inst.output_on[2] = 'on'

        print('Trigger type:{}'.format(inst.trig_output))


        #inst.timed_delay[1] = 100e-9 #ns
        #inst.timed_delay[2] = 250e-9 #ns

        print('Channel 1 timed_delay:{}'.format(inst.timed_delay[1]))
        print('Channel 2 timed_delay:{}'.format(inst.timed_delay[2]))
        # set up 3 different segments
        # start at segment 1 and loop through all 3 segments

        T_init = 13200
        T_rabi_max = 224
        T_readout = 13200

        # Segment 1 - laser initializes spin for T_init
        inst.dig_patt_length[1] = T_init
        #print('Segment {} length:{}'.format(1, inst.dig_patt_length[1]))

        # Segment 1 - RF off
        inst.dig_patt_type[(1, 1)] = 'low'
        #print('Channel {}, segment {} type:{}'.format(1, 1, inst.dig_patt_type[(1, 1)]))

        # Segment 1 - laser on, initializing spin
        inst.dig_patt_type[(2, 1)] = 'high'
        #print('Channel {}, segment {} type:{}'.format(2, 1, inst.dig_patt_type[(2, 1)]))


        # Segment 2 - apply variable length RF pulse

        # 2 rf is on for variable time tau_rf
        inst.dig_patt_length[2] = T_rabi_max
        #print('Segment {} length:{}'.format(2, inst.dig_patt_length[2]))

        inst.dig_patt_type[(1,2)] = 'data'
        #print('Channel {}, segment {} type:{}'.format(1, 2, inst.dig_patt_type[(1, 2)]))

        # Set up segment 2 RF - initial point is with no RF on
        inst.segment_data[(1,2)] = [np.zeros(T_rabi_max, dtype='int')]

        #print('Channel {}, segment {} data:'.format(1, 2, inst.segment_data[(1,2)]))

        # Segment 2 - laser is off
        inst.dig_patt_type[(2, 2)] = 'low'
        #print('Channel {}, segment {} type:{}'.format(2, 2, inst.dig_patt_type[(2, 2)]))

        # Segment 3 - laser reads out, initializes, waits, reads out
        inst.dig_patt_length[3] = T_rabi_max + T_init + 2 * T_readout
        #print('Segment {} length:{}'.format(3, inst.dig_patt_length[3]))

        # Segment 3 - RF is always off
        inst.dig_patt_type[(1, 3)] = 'low'
        #print('Channel {}, segment {} type:{}'.format(1, 3, inst.dig_patt_type[(1, 3)]))

        # Segment 3 - laser initializes, waits, reads out
        inst.dig_patt_type[(2, 3)] = 'data'
        #print('Channel {}, segment {} type:{}'.format(2, 3, inst.dig_patt_type[(2, 3)]))

        readout1 = [np.hstack((np.ones(T_readout + T_init, dtype='int'), np.zeros(T_rabi_max, dtype='int')))]

        print(inst.dig_patt_length)
        #print(readout[0].shape)

        inst.segment_data[(2,3)] = readout1 #[np.hstack((np.ones(T_readout + T_init, dtype='int'), np.zeros(T_rabi_max, dtype='int'), np.ones(T_readout, dtype='int')))]

        # Segment 3 - RF is always off
        inst.dig_patt_type[(1, 3)] = 'low'
        #print('Channel {}, segment {} type:{}'.format(1, 3, inst.dig_patt_type[(1, 3)]))

        # Segment 3 - laser initializes, waits, reads out
        inst.dig_patt_type[(2, 3)] = 'data'
        #print('Channel {}, segment {} type:{}'.format(2, 3, inst.dig_patt_type[(2, 3)]))

        inst.dig_patt_type[(1, 4)] = 'low'

        inst.dig_patt_type[(2, 4)] = 'data'

        readout2 = [np.hstack((np.ones(T_readout, dtype='int')))]

        print(inst.dig_patt_length)
        #print(readout[0].shape)

        inst.segment_data[(2,4)] = readout2 #[np.hstack((np.ones(T_readout + T_init, dtype='int'), np.zeros(T_rabi_max, dtype='int'), np.ones(T_readout, dtype='int')))]





        # sets PPG to loop through segments 1-3 repeatedly
        inst.loop_start = 1
        inst.loop_length = 3

        print('Trigger source?: {}'.format(inst.trigger_source))

        inst.trigger_source = 'internal'
        print('Trigger source?: {}'.format(inst.trigger_source))


        print('trigger mode:{}'.format(inst.trig_mode))


        print(inst.trig_output)
        print('trigger position:{}'.format(inst.trig_pos))

        #from time import sleep

        #sleep(10)

        #inst.write('ARM:SOUR MAN')
        #inst.write('ARM:MODE STAR')

        #inst.write('DIG:PATT:LOOP:INF ON')
        #inst.write('DIG:PATT:INST:STAR SEGM1')
        #inst.write('DIG:PATT:LOOP:LENG 3')

        #inst.preview_wfm()
        #
        # data = np.random.randint(2, size=100, dtype=np.uint8)
        # encoded = inst.encode_data(data)
        #
        # print('Data:{}'.format(data))
        # print('Encoded:{}'.format(encoded))
        #
        # decoded = inst.decode_data('{}'.format(encoded))
        # print('Decoded:{}'.format(decoded))
        # for segment in segments:
        #
        #     segment_length = inst.dig_patt_length[segment]
        #
        #     for channel in [1,2]:
        #
        #         v_high = inst.volt_high[channel]
        #         v_low = inst.volt_low[channel]
        #
        #         segment_type = inst.dig_patt_type[(channel, segment)]
        #
        #         if segment_data == 'data':
        #             pass
        #             #data = inst.segment_data[(channel, segment)]
        #
        #
        #         #print(segment)
        #         #print(inst.segment_data[(channel, segment)])
        #
        #     #print(data)
        #     plt.plot(data)
        #     plt.xlabel('Time')
        #     plt.ylabel('Voltage')
        #     plt.show()


        # figure out these numbers!
        T_init = 1000 #us
        T_gap = 50 #us
        T_readout = 1000 #us

        tau_rf = 100*1e-3 #100 ns


        # program pattern of RF on for tau after T_init + T_gap / 2 - tau_rf




        # figure out how to vary tau_rf + have it still placed appropriately

    def pulse_odmr_setup(inst):
        """
        Sets up PPG output channels for pulsed ODMR measurements.
        """
        print('Identification: {}'.format(inst.idn))
        inst.reset() #initializes default parameters for clean setup
        inst.display = 'off'
        print('Display off?: {}'.format(inst.display))

        inst.pattern_mode = 'on'
        print('Digital pattern mode on?:{}'.format(inst.pattern_mode))

        #inst.arm_source = 'immediate' # sets continuous operation
        #print('Arm source immediate?: {}'.format(inst.arm_source))

        inst.frequency = 660e6
        # output TTL pulses for RF switch on channel 1
        # TTL pulses should be between 0 (low) and 2.5 (high) volts
        # so set up channel 1 output like this
        inst.volt_low[1] = 0.0
        inst.volt_high[1] = 2.5
        print('MW TTL high voltage, should be 2.5 V:{}'.format(inst.volt_high[1]))
        print('MW TTL low voltage, should be 0 V:{}'.format(inst.volt_low[1]))

        inst.data_format[1] = 'nrz'
        inst.output_on[1] = 'on'
        inst.comp_output_on[1] = 'on' #for scope viewing

        # set up laser channel
        inst.volt_low[2] = 0.0
        inst.volt_high[2] = 1.0
        print('AOM high voltage, should be 1.0 V:{}'.format(inst.volt_high[2]))
        print('AOM Low voltage, should be 0 V:{}'.format(inst.volt_low[2]))

        inst.data_format[2] = 'nrz'
        inst.output_on[2] = 'on'
        inst.comp_output_on[2] = 'on' #for scope viewing


        print('Trigger type:{}'.format(inst.trig_output))

        #inst.timed_delay[1] = 100e-9 #ns
        #inst.timed_delay[2] = 250e-9 #ns

        print('Channel 1 timed_delay:{}'.format(inst.timed_delay[1]))
        print('Channel 2 timed_delay:{}'.format(inst.timed_delay[2]))

    def rabi_step(inst, step_number):
        """
        Sets up next waveform point for Rabi
        """
        T_rabi_max = 4096

        data = [np.hstack((np.ones(step_number, dtype='int'), np.zeros(T_rabi_max - step_number, dtype='int')))]

        encoded = inst.encode_data(data[0])

        seg_num = 3
        channel = 1

        inst.write('DIG:PATT:SEGM{}:DATA{} {}'.format(seg_num, channel, encoded))


    def ramsey_setup(inst, ramsey_params, pi_pulse_len):

        T_ramsey_max = 6144
        T_init = 336*5 #~2.5us
        T_gap = 16*15 # segment 2, everything off
        T_readout = 336*5 #~2.5us

        pi_pulse = np.ones(pi_pulse_len, dtype='int')
        pad = np.zeros(T_ramsey_max - pi_pulse_len, dtype='int')


        # Segment 1 - RF off, laser on to initialize
        inst.dig_patt_length[1] = T_init
        inst.dig_patt_type[(1, 1)] = 'low'
        inst.dig_patt_type[(2, 1)] = 'high'

        # Segment 2 - gap, everything is off
        inst.dig_patt_length[2] = T_gap
        inst.dig_patt_type[(1,2)] = 'low'
        inst.dig_patt_type[(2,2)] = 'low'

        # Segment 3 - laser is off, variable length RF pulse
        inst.dig_patt_length[3] = T_ramsey_max
        inst.dig_patt_type[(2, 3)] = 'low'
        inst.dig_patt_type[(1,3)] = 'data'

        # Set up segment 2 RF - initial point is pi pulse w/o separation


        inst.segment_data[(1,3)] = [np.hstack((pi_pulse, pad))]

        # Segment 4 - laser reads out
        inst.dig_patt_length[4] = T_readout
        # Segment 4 - RF is always off, laser on
        inst.dig_patt_type[(1, 4)] = 'low'
        inst.dig_patt_type[(2, 4)] = 'high'


    def ramsey_step(inst, ramsey_params, pi_pulse_len, tau):

        T_ramsey_max = 6144


        pi_pulse = np.ones(pi_pulse_len, dtype='int')

        pi2_pulse = np.ones(int(pi_pulse_len/2), dtype='int')

        delay = np.zeros(tau, dtype='int')

        pad = np.zeros((T_ramsey_max - (tau + pi_pulse_len)), dtype='int')

        data = np.hstack((pi2_pulse, delay, pi2_pulse, pad))

        seg_num = 3
        channel = 1

        encoded = inst.encode_data(data)

        inst.write('DIG:PATT:SEGM{}:DATA{} {}'.format(seg_num, channel, encoded))



    def hahn_setup(inst, hahn_params):

        freq = hahn_params['ppg_freq']
        inst.frequency = freq

        conversion = freq / 660e6

        T_hahn_max = hahn_params['T_hahn_max']
        T_init = int(hahn_params['T_init'] * conversion)
        T_gap = int(hahn_params['T_gap'] * conversion)
        T_readout = int(hahn_params['T_readout'] * conversion)
        tau_min = int(hahn_params['tau_min'] * conversion)


        pi_pulse_len = 2

        pad = np.zeros(int((T_hahn_max - 2*pi_pulse_len - tau_min)/2.0), dtype='int')


        pi_pulse = np.ones(pi_pulse_len, dtype='int')
        pi2_pulse = np.ones(int(pi_pulse_len/2), dtype='int')
        tau2 = np.zeros(int(tau_min/2), dtype='int')

        # Segment 1 - RF off, laser on to initialize
        inst.dig_patt_length[1] = T_init
        inst.dig_patt_type[(1, 1)] = 'low'
        inst.dig_patt_type[(2, 1)] = 'high'

        # Segment 2 - gap, everything is off
        inst.dig_patt_length[2] = T_gap
        inst.dig_patt_type[(1,2)] = 'low'
        inst.dig_patt_type[(2,2)] = 'low'

        # Segment 3 - laser is off, pi/2, tau, pi, -pi/2 pulses
        inst.dig_patt_length[3] = T_hahn_max
        inst.dig_patt_type[(2, 3)] = 'low'

        inst.dig_patt_type[(1,3)] = 'data'


        hahn_data = np.hstack((pad, pi2_pulse, tau2, pi_pulse, tau2,
                               pi2_pulse, pad))

        print(T_hahn_max)
        print(hahn_data.shape)


        inst.segment_data[(1,3)] = [hahn_data]


        # Segment 4 - laser reads out
        inst.dig_patt_length[4] = T_readout
        # Segment 4 - RF is always off, laser on
        inst.dig_patt_type[(1, 4)] = 'low'
        inst.dig_patt_type[(2, 4)] = 'high'




    def hahn_step(inst, hahn_params, tau):

        T_hahn_max = hahn_params['T_hahn_max']
        pi_pulse_len = hahn_params['pi_pulse_len']

        pi_pulse = np.ones(pi_pulse_len, dtype='int')
        pi2_pulse = np.ones(int(pi_pulse_len/2), dtype='int')
        tau2 = np.zeros(int(tau/2), dtype='int')

        pad = np.zeros(int((T_hahn_max - 2*pi_pulse_len - tau)/2.0), dtype='int')

        hahn_data = np.hstack((pad, pi2_pulse, tau2, pi_pulse, tau2, pi2_pulse, pad))

        print(hahn_data.shape)

        seg_num = 3
        channel = 1

        encoded = inst.encode_data(hahn_data)

        inst.write('DIG:PATT:SEGM{}:DATA{} {}'.format(seg_num, channel, encoded))

        inst.segment_data[(channel, seg_num)]

        return

    def rabi_setup(inst):
        """
        Sets up pulse sequence for doing Rabi
        """
        T_init = 336*6 #500 ns
        T_gap = 16*15 # segment 2, everything off
        T_rabi_max = 4096

        T_readout = 336*6 # 500 ns


        # Segment 1 - RF off, laser on to initialize
        inst.dig_patt_length[1] = T_init
        inst.dig_patt_type[(1, 1)] = 'low'
        inst.dig_patt_type[(2, 1)] = 'high'

        # Segment 2 - gap, everything is off
        inst.dig_patt_length[2] = T_gap
        inst.dig_patt_type[(1,2)] = 'low'
        inst.dig_patt_type[(2,2)] = 'low'

        # Segment 3 - laser is off, variable length RF pulse
        inst.dig_patt_length[3] = T_rabi_max
        inst.dig_patt_type[(2, 3)] = 'low'
        inst.dig_patt_type[(1,3)] = 'data'

        # Set up segment 2 RF - initial point is with no RF on
        inst.segment_data[(1,3)] = [np.zeros(T_rabi_max, dtype='int')]

        # Segment 4 - laser reads out
        inst.dig_patt_length[4] = T_readout
        # Segment 4 - RF is always off, laser on
        inst.dig_patt_type[(1, 4)] = 'low'
        inst.dig_patt_type[(2, 4)] = 'high'

        print('Trigger source?: {}'.format(inst.trigger_source))

        inst.trigger_source = 'internal'
        print('Trigger source?: {}'.format(inst.trigger_source))


        print('trigger mode:{}'.format(inst.trig_mode))


        print(inst.trig_output)
        print('trigger position:{}'.format(inst.trig_pos))


    def calibrate_pi_setup(inst, downconversion_rate):

        inst.frequency = 660e6 / downconversion_rate
        print(inst.frequency)


        T_init = int(336*6 / downconversion_rate) #500 ns
        T_gap = int(16*15 / downconversion_rate) # segment 2, everything off
        T_rabi_max = int(4096 / downconversion_rate)

        T_readout = int(336*6 / downconversion_rate) # 500 ns


        # Segment 1 - RF off, laser on to initialize
        inst.dig_patt_length[1] = T_init
        inst.dig_patt_type[(1, 1)] = 'low'
        inst.dig_patt_type[(2, 1)] = 'high'

        # Segment 2 - gap, everything is off
        inst.dig_patt_length[2] = T_gap
        inst.dig_patt_type[(1,2)] = 'low'
        inst.dig_patt_type[(2,2)] = 'low'

        # Segment 3 - laser is off, variable length RF pulse
        inst.dig_patt_length[3] = T_rabi_max
        inst.dig_patt_type[(2, 3)] = 'low'
        inst.dig_patt_type[(1,3)] = 'data'

        # Set up segment 2 RF - initial point is with no RF on
        num_pulses = 5

        data = np.hstack((np.ones(2*num_pulses, dtype='int'), np.zeros(T_rabi_max - 2*num_pulses, dtype='int')))

        inst.segment_data[(1,3)] = [data]

        # Segment 4 - laser reads out
        inst.dig_patt_length[4] = T_readout
        # Segment 4 - RF is always off, laser on
        inst.dig_patt_type[(1, 4)] = 'low'
        inst.dig_patt_type[(2, 4)] = 'high'

        inst.trigger_source = 'internal'

        return




if __name__ == '__main__':

    test_wfm = True
    test_rabi = False
    test_comm = False

    gpib_addr = 10

    with Ag81130A('GPIB0::{}::INSTR'.format(gpib_addr)) as inst:

        #inst.preview_wfm()

        if test_wfm:
            inst.odmr_waveform()

        elif test_rabi:

            rabi_waveform_test(inst)

        elif test_comm:

            print('Identification: {}'.format(inst.idn))
            inst.reset()
            print('Display: {}'.format(inst.display))
            inst.display = 'off'
            print('Display: {}'.format(inst.display))

            print('Digital pattern mode: {}'.format(inst.pattern_mode))
            inst.pattern_mode = 'on'
            print('Digital pattern mode: {}'.format(inst.pattern_mode))

            print('Arm source: {}'.format(inst.arm_source))
            inst.arm_source = 'manual'
            print('Arm source: {}'.format(inst.arm_source))
            inst.arm_source = 'immediate'
            print('Arm source: {}'.format(inst.arm_source))


            for segment in range(1,5):
                print('Segment {} length:{}'.format(segment, inst.dig_patt_length[segment]))
                inst.dig_patt_length[segment] = 100
                print('Segment {} length:{}'.format(segment, inst.dig_patt_length[segment]))



            for channel in [1,2]:
                print('Channel {} high:{}V'.format(channel, inst.volt_high[channel]))
                inst.volt_high[channel] = 3.0
                print('Channel {} high:{}V'.format(channel, inst.volt_high[channel]))
                inst.volt_high[channel] = 2.5
                print('Channel {} high:{}V'.format(channel, inst.volt_high[channel]))

                print('Channel {} low:{}V'.format(channel, inst.volt_low[channel]))
                inst.volt_low[channel] = -1.0
                print('Channel {} low:{}V'.format(channel, inst.volt_low[channel]))
                inst.volt_low[channel] = 0.0
                print('Channel {} low:{}V'.format(channel, inst.volt_low[channel]))

                for segment in range(1,5):

                    inst.dig_patt_type[(channel, segment)] = 'high'
                    print('Channel {}, segment {} type:{}'.format(channel, segment, inst.dig_patt_type[(channel, segment)]))

                print('Channel format:{}'.format(inst.data_format[channel]))
                inst.data_format[channel] = 'nrz'
                print('Channel format:{}'.format(inst.data_format[channel]))

                print('Channel output:{}'.format(inst.output_on[channel]))
                inst.output_on[channel] = 'on'
                print('Channel output:{}'.format(inst.output_on[channel]))
